# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.source_policy import apply_stream_source
from contrast.agent.policy import patch_manager
from contrast.agent.middlewares.route_coverage import common

from contrast.utils.patch_utils import (
    build_and_apply_patch,
    unregister_module_patcher,
    wrap_and_watermark,
    register_module_patcher,
)
from contrast.utils.decorators import fail_quietly
from contrast_vendor import structlog as logging


FALCON_MULTIPART_MIDDLEWARE_MODULE = "falcon_multipart.middleware"
FALCON_MULTIPART_MIDDLEWARE_CLASS = "MultipartMiddleware"
FALCON_MODULE = "falcon"
FALCON_ASGI_MODULE = "falcon.asgi"

logger = logging.getLogger("contrast")


def build_parse_patch(orig_method, _):
    def parse_patch(wrapped, instance, args, kwargs):
        """
        Deadzone for falcon_multipart.middleware.MultipartMiddleware.parse

        This method can cause an enormous amount of propagation to occur, which can
        cause requests to be extremely slow. This is because this middleware parses all
        of the multipart form data into memory on the request object. We deadzone this
        propagation in order to greatly improve performance. The parse_field method
        patch below ensures that we still track the relevant data.

        This method gets called before MultipartMiddleware.parse_field in
        MultipartMiddleware.process_request.
        """
        del instance

        with scope.contrast_scope():
            return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_method, parse_patch)


def build_parse_field_patch(orig_method, _):
    def parse_field_patch(wrapped, instance, args, kwargs):
        """
        Deadzone and source tracker for falcon_multipart.middleware.MultipartMiddleware

        First we deadzone the call to the original method in order to prevent
        unnecessary propagation and improve performance. Next, we create sources for
        the result so that we don't miss any input data.

        This method gets called after MultipartMiddleware.parse in
        MultipartMiddleware.process_request.
        """
        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)

        apply_stream_source("parse_field", result, instance, result, args, kwargs)

        return result

    return wrap_and_watermark(orig_method, parse_field_patch)


def patch_falcon_multipart(module):
    middleware_cls = getattr(module, FALCON_MULTIPART_MIDDLEWARE_CLASS, None)
    if middleware_cls is None:
        return

    build_and_apply_patch(middleware_cls, "parse", build_parse_patch)
    build_and_apply_patch(middleware_cls, "parse_field", build_parse_field_patch)


def build_call_patch(orig_func, patch_policy):
    """
    Patch for falcon.App.__call__

    This is how we perform route discovery and observation for Falcon.
    """
    del patch_policy

    def call_patch(wrapped, instance, args, kwargs):
        do_falcon_first_request_analysis(instance)
        try:
            result = wrapped(*args, **kwargs)
        finally:
            do_falcon_routes_observation(instance, args)

        return result

    return wrap_and_watermark(orig_func, call_patch)


@fail_quietly("unable to perform Falcon first request analysis discovery")
@scope.contrast_scope()
def do_falcon_first_request_analysis(falcon_instance):
    from contrast.agent.middlewares.route_coverage.falcon_routes import (
        create_falcon_routes,
    )

    from contrast.agent import agent_state

    if not agent_state.is_first_request():
        return

    common.handle_route_discovery("falcon", create_falcon_routes, (falcon_instance,))
    do_config_scanning(logger, falcon_instance)


@fail_quietly("unable to perform Falcon route observation")
@scope.contrast_scope()
def do_falcon_routes_observation(falcon_app_instance, request_args):
    from contrast.agent.middlewares.route_coverage.falcon_routes import (
        get_falcon_signature_and_template,
    )

    context = contrast.REQUEST_CONTEXT.get()

    if context is None:
        return
    if getattr(falcon_app_instance, "_ASGI", False):
        request_path = request_args[0]["path"]
        request_method = request_args[0]["method"]
    else:
        request_path = request_args[0]["PATH_INFO"]
        request_method = request_args[0]["REQUEST_METHOD"]

    context.signature, context.path_template = get_falcon_signature_and_template(
        request_path,
        falcon_app_instance,
        request_method,
    )
    logger.debug(
        "Found Falcon view function",
        signature=context.signature,
        path_template=context.path_template,
    )


@fail_quietly("failed to apply Falcon config scanning rules")
def do_config_scanning(logger, falcon_app_instance):
    from contrast.agent.assess.rules.config.falcon_secure_flag_rule import (
        FalconSecureFlagRule,
    )

    logger.debug("applying falcon config scanning rules")

    for rule in [
        FalconSecureFlagRule(),
    ]:
        rule.apply(falcon_app_instance.resp_options)


def patch_falcon(falcon_module):
    # If the runner is in use, then CommonMiddlewarePatch has already
    # patched __call__ for foundational middleware operations. We want
    # both patches to be applied, so we force this one.
    build_and_apply_patch(falcon_module.App, "__call__", build_call_patch, force=True)


def patch_falcon_asgi(falcon_asgi_module):
    build_and_apply_patch(
        falcon_asgi_module.App, "__call__", build_call_patch, force=True
    )


def register_patches():
    register_module_patcher(patch_falcon, FALCON_MODULE)
    register_module_patcher(patch_falcon_asgi, FALCON_ASGI_MODULE)

    register_module_patcher(patch_falcon_multipart, FALCON_MULTIPART_MIDDLEWARE_MODULE)


def reverse_patches():
    unregister_module_patcher(FALCON_MODULE)
    if falcon_module := sys.modules.get(FALCON_MODULE):
        patch_manager.reverse_patches_by_owner(falcon_module.App)

    unregister_module_patcher(FALCON_ASGI_MODULE)
    if falcon_asgi_module := sys.modules.get(FALCON_ASGI_MODULE):
        patch_manager.reverse_patches_by_owner(falcon_asgi_module.app.App)
        patch_manager.reverse_patches_by_owner(falcon_asgi_module.App)
        patch_manager.reverse_patches_by_owner(falcon_asgi_module)

    unregister_module_patcher(FALCON_MULTIPART_MIDDLEWARE_MODULE)
    if (
        falcon_multipart_middleware_module := sys.modules.get(
            FALCON_MULTIPART_MIDDLEWARE_MODULE
        )
    ) and (
        middleware_cls := getattr(
            falcon_multipart_middleware_module, FALCON_MULTIPART_MIDDLEWARE_CLASS
        )
    ):
        patch_manager.reverse_patches_by_owner(middleware_cls)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
This module contains core instrumentation for patching builtin methods.

These are the "pure python string patches" that are applied to strlike types using
set_attr_on_type. They are called directly when the relevant strlike method is called -
our C extension is not called first to dispatch to them.
"""

import _io
from collections import UserString
import contrast
from contrast.agent import scope
from contrast.agent.assess.utils import is_tracked
from contrast.agent.policy import patch_manager, registry
from contrast.agent.assess.policy.propagation_policy import (
    PROPAGATOR_ACTIONS,
    propagate_stream,
    track_copy_without_new_event,
)
from contrast.agent.assess.policy.propagators import FormatPropagator, JoinPropagator
from contrast.agent.assess.policy.preshift import Preshift
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark
from contrast.assess_extensions import smart_setattr


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_bytearray_join_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def bytearray_join_patch(wrapped, instance, args, kwargs):
        # Since we need to make reference to the input multiple times, convert the
        # first argument to a list and use that instead. This prevents any iterators
        # from being exhausted before we can make use of them in propagation.
        # For bytearray.join, args == (list_or_iterator_of_things_to_join,...)
        # Note that this is different from the C hooks for other join methods. In
        # those cases, the PyObject *args argument corresponds to just the list or
        # iterator itself, in contrast to a tuple that contains that list or
        # iterator. (Got that straight?)
        if scope.in_contrast_or_propagation_scope():
            return wrapped(*args, **kwargs)

        args_list = [list(args[0])] + list(args[1:])
        result = wrapped(*args_list, **kwargs)

        with scope.contrast_scope():
            try:
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, args_list, kwargs)
                propagator = JoinPropagator(node, preshift, result)
                if propagator.needs_propagation:
                    propagator.track_and_propagate(result)
                    context.propagated()
            except Exception as ex:
                logger.debug("failed to propagate bytearray.join", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, bytearray_join_patch)


def build_strtype_join_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def strtype_join_patch(wrapped, instance, args, kwargs):
        if scope.in_contrast_or_propagation_scope():
            return wrapped(*args, **kwargs)

        args_list = list(args[0])

        with scope.propagation_scope():
            result = wrapped(args_list, **kwargs)

        with scope.contrast_scope():
            try:
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, (args_list,), kwargs)
                propagator = JoinPropagator(node, preshift, result)
                if propagator.needs_propagation:
                    propagator.track_and_propagate(result)
                    context.propagated()
            except Exception as ex:
                logger.debug("failed to propagate join", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, strtype_join_patch)


def build_str_format_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def str_format_patch(wrapped, instance, args, kwargs):
        """
        Propagation hook for str.format

        This hook is a special case because we need to enable some propagation to occur
        while we evaluate whether to propagate this particular event. With the current
        general hook infrastructure, this is not possible, so we need to account for it
        here. Eventually it may be possible to fit this back into the more general
        infrastructure if we overhaul the way that scope works.
        """
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        try:
            with scope.contrast_scope():
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, args, kwargs)
                propagator = FormatPropagator(node, preshift, result)

            # This evaluation must not occur in scope. This is what enables us
            # to perform any conversions from object to __str__ or __repr__,
            # while allowing propagation to occur through those methods if
            # necessary.
            if propagator.needs_propagation:
                with scope.contrast_scope():
                    propagator.track_and_propagate(result)
                    context.propagated()
        except Exception as ex:
            with scope.propagation_scope():
                logger.debug("failed to propagate str.format", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, str_format_patch)


def build_str_formatmap_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def str_formatmap_patch(wrapped, instance, args, kwargs):
        """
        Propagation hook for str.format_map

        This hook is a special case because we need to enable some propagation to occur
        while we evaluate whether to propagate this particular event. With the current
        general hook infrastructure, this is not possible, so we need to account for it
        here. Eventually it may be possible to fit this back into the more general
        infrastructure if we overhaul the way that scope works.
        """
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope() or not args:
            return result

        try:
            with scope.contrast_scope():
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, (), args[0])
                propagator = FormatPropagator(node, preshift, result)

            # This evaluation must not occur in scope. This is what enables us
            # to perform any conversions from object to __str__ or __repr__,
            # while allowing propagation to occur through those methods if
            # necessary.
            if propagator.needs_propagation:
                with scope.contrast_scope():
                    propagator.track_and_propagate(result)
                    context.propagated()
        except Exception as ex:
            with scope.propagation_scope():
                logger.debug("failed to propagate str.format", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, str_formatmap_patch)


def build_generic_strtype_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]
    propagator_class = PROPAGATOR_ACTIONS.get(node.action)

    def str_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        # This special case does not apply to bytearrays
        if result is instance:
            return result

        if scope.in_contrast_or_propagation_scope():
            return result

        with scope.contrast_scope():
            try:
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, args, kwargs)
                propagator = propagator_class(node, preshift, result)

                if propagator.needs_propagation:
                    propagator.track_and_propagate(result)
                    context.propagated()
            except Exception as ex:
                name = orig_method.__class__.__name__
                logger.debug(
                    "failed to propagate %s.%s", name, orig_method.__name__, exc_info=ex
                )

        return result

    return wrap_and_watermark(orig_method, str_patch)


def build_track_without_new_event_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]
    propagator_class = PROPAGATOR_ACTIONS.get(node.action)

    def str_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        # This special case applies to all bytearray methods and all .translate methods
        if result == instance:
            track_copy_without_new_event(result, instance)
            return result

        with scope.contrast_scope():
            try:
                if (
                    context := contrast.REQUEST_CONTEXT.get()
                ) is None or context.stop_propagation:
                    return result

                preshift = Preshift(instance, args, kwargs)
                propagator = propagator_class(node, preshift, result)

                if propagator.needs_propagation:
                    propagator.track_and_propagate(result)
                    context.propagated()
            except Exception as ex:
                name = orig_method.__class__.__name__
                logger.debug(
                    "failed to propagate %s.%s", name, orig_method.__name__, exc_info=ex
                )

        return result

    return wrap_and_watermark(orig_method, str_patch)


def build_generic_stream_patch(orig_method, patch_policy):
    del patch_policy
    method_name = orig_method.__name__

    def stream_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        if scope.in_contrast_or_propagation_scope():
            return result

        if not (
            getattr(instance, "cs__tracked", False)
            or getattr(instance, "cs__source", False)
        ):
            return result

        if is_tracked(result):
            return result

        with scope.propagation_scope(), scope.contrast_scope():
            if (
                context := contrast.REQUEST_CONTEXT.get()
            ) is None or context.stop_propagation:
                return result

            propagate_stream(method_name, result, instance, result, args, kwargs)
            context.propagated()

        return result

    return wrap_and_watermark(orig_method, stream_patch)


def build_and_apply_str_patch(owner, method_name, patch_builder):
    orig_method = getattr(owner, method_name)

    policy_method_name = (
        "formatmap" if orig_method.__name__ == "format_map" else orig_method.__name__
    )

    patch_policy = registry.get_policy_by_name("builtins.str." + policy_method_name)
    patch = patch_builder(orig_method, patch_policy)

    patch_manager.patch(owner, method_name, patch)


def property_getter(self):
    return contrast.STRING_TRACKER.get(self, None)


def property_setter(self, value):
    contrast.STRING_TRACKER.update_properties(self, value)


def enable_str_properties():
    strprop = property(fget=property_getter, fset=property_setter)

    smart_setattr(str, "cs__properties", strprop)
    smart_setattr(bytes, "cs__properties", strprop)
    smart_setattr(bytearray, "cs__properties", strprop)

    for stream_type in (_io.StringIO, _io.BytesIO):
        smart_setattr(stream_type, "cs__tracked", False)
        smart_setattr(stream_type, "cs__source", False)
        smart_setattr(stream_type, "cs__properties", None)
        smart_setattr(stream_type, "cs__source_event", None)
        smart_setattr(stream_type, "cs__source_type", None)
        smart_setattr(stream_type, "cs__source_tags", None)

    # IPython assumes that UserString has the same attributes as str,
    # and raises AttributeErrors if that is not the case. So any
    # properties we add to str should also be added to UserString.
    # UserString will dispatch to str methods on self.data though, so
    # we don't need to actually add a strprop onto the type.
    smart_setattr(UserString, "cs__properties", None)


def patch_strtype_method(strtype, method_name):
    if patch_manager.is_patched(getattr(strtype, method_name)):
        return

    if method_name == "join":
        builder = (
            build_bytearray_join_patch
            if strtype is bytearray
            else build_strtype_join_patch
        )
    elif method_name == "format":
        builder = build_str_format_patch
    elif method_name == "format_map":
        builder = build_str_formatmap_patch
    elif method_name == "translate" and strtype is str:
        builder = build_track_without_new_event_patch
    else:
        builder = (
            build_track_without_new_event_patch
            if strtype is bytearray
            else build_generic_strtype_patch
        )

    build_and_apply_str_patch(strtype, method_name, patch_builder=builder)


def patch_stream_method(stream_type, method_name):
    build_and_apply_patch(stream_type, method_name, build_generic_stream_patch)


def unpatch_strtype_and_stream_methods():
    """
    Replace all patched strtype and stream methods with the original implementation.

    NOTE: repeatedly enabling and disabling these patches in the same python process can
    cause tp_version_tag to reach its maximum allowed value in the interpreter. This
    shouldn't be an issue, but it has been known to cause problems in the past. Minimize
    the use of this function.
    """
    patch_manager.reverse_patches_by_owner(str)
    patch_manager.reverse_patches_by_owner(bytes)
    patch_manager.reverse_patches_by_owner(bytearray)

    patch_manager.reverse_patches_by_owner(_io.StringIO)
    patch_manager.reverse_patches_by_owner(_io.BytesIO)
    patch_manager.reverse_patches_by_owner(_io._IOBase)

    patch_manager.reverse_patches_by_owner(UserString)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Controller for global scope state

Basically we use scoping to prevent us from assessing our own code. Scope
improves performance but it also prevents us from accidentally recursing
inside our analysis code. For example, we don't want to inadvertently cause
string propagation events while we're doing string building for reporting
purposes.
"""

import contextlib
from contextvars import ContextVar
import functools

CONTRAST_SCOPE = ContextVar("contrast_scope", default=0)
PROPAGATION_SCOPE = ContextVar("propagation_scope", default=0)
TRIGGER_SCOPE = ContextVar("trigger_scope", default=0)
OBSERVE_SCOPE = ContextVar("observe_scope", default=0)


def current_scope():
    return (
        CONTRAST_SCOPE.get(),
        PROPAGATION_SCOPE.get(),
        TRIGGER_SCOPE.get(),
        OBSERVE_SCOPE.get(),
    )


def set_scope(
    contrast_scope: int, propagation_scope: int, trigger_scope: int, observe_scope: int
):
    CONTRAST_SCOPE.set(contrast_scope)
    PROPAGATION_SCOPE.set(propagation_scope)
    TRIGGER_SCOPE.set(trigger_scope)
    OBSERVE_SCOPE.set(observe_scope)


def in_contrast_scope() -> bool:
    return CONTRAST_SCOPE.get() > 0


def in_propagation_scope() -> bool:
    return PROPAGATION_SCOPE.get() > 0


def in_trigger_scope() -> bool:
    return TRIGGER_SCOPE.get() > 0


def in_observe_scope() -> bool:
    return OBSERVE_SCOPE.get() > 0


def _build_scope_manager(scope: ContextVar):
    class _ScopeManager:
        def __enter__(self):
            # TODO: PYT-3433 ideally we'd use a token + reset() here, but this causes
            # LookupErrors in rare circumstances that we don't currently understand
            scope.set(scope.get() + 1)

        def __exit__(self, exc_type, exc_val, exc_tb):
            scope.set(scope.get() - 1)

        def __call__(self, f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                scope.set(scope.get() + 1)
                try:
                    return f(*args, **kwargs)
                finally:
                    scope.set(scope.get() - 1)

            return wrapper

    return _ScopeManager


contrast_scope = _build_scope_manager(CONTRAST_SCOPE)
propagation_scope = _build_scope_manager(PROPAGATION_SCOPE)
trigger_scope = _build_scope_manager(TRIGGER_SCOPE)
observe_scope = _build_scope_manager(OBSERVE_SCOPE)

###################################################
# Convenience functions not needed for all scopes #
###################################################


def in_contrast_or_propagation_scope():
    """Indicates we are in either contrast scope or propagation scope"""
    return in_contrast_scope() or in_propagation_scope()


@contextlib.contextmanager
def pop_contrast_scope():
    """
    Context manager that pops contrast scope and restores the previous scope level
    when it exits.
    """
    CONTRAST_SCOPE.set(CONTRAST_SCOPE.get() - 1)
    try:
        yield
    finally:
        CONTRAST_SCOPE.set(CONTRAST_SCOPE.get() + 1)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Helpers that make development with *structlog* more pleasant.

See also the narrative documentation in `console-output`.
"""

from __future__ import annotations

import shutil
import sys
import warnings

from dataclasses import dataclass
from io import StringIO
from types import ModuleType
from typing import (
    Any,
    Callable,
    Literal,
    Protocol,
    Sequence,
    TextIO,
    Type,
    Union,
    cast,
)

from ._frames import _format_exception
from .processors import _figure_out_exc_info
from .typing import EventDict, ExceptionRenderer, ExcInfo, WrappedLogger


try:
    import colorama
except ImportError:
    colorama = None

try:
    import better_exceptions
except ImportError:
    better_exceptions = None

try:
    import rich

    from rich.console import Console
    from rich.traceback import Traceback
except ImportError:
    rich = None  # type: ignore[assignment]


__all__ = [
    "ConsoleRenderer",
    "plain_traceback",
    "rich_traceback",
    "better_traceback",
]

_IS_WINDOWS = sys.platform == "win32"

_MISSING = "{who} requires the {package} package installed.  "
_EVENT_WIDTH = 30  # pad the event name to so many characters


def _pad(s: str, length: int) -> str:
    """
    Pads *s* to length *length*.
    """
    missing = length - len(s)

    return s + " " * (max(0, missing))


if colorama is not None:
    RESET_ALL = colorama.Style.RESET_ALL
    BRIGHT = colorama.Style.BRIGHT
    DIM = colorama.Style.DIM
    RED = colorama.Fore.RED
    BLUE = colorama.Fore.BLUE
    CYAN = colorama.Fore.CYAN
    MAGENTA = colorama.Fore.MAGENTA
    YELLOW = colorama.Fore.YELLOW
    GREEN = colorama.Fore.GREEN
    RED_BACK = colorama.Back.RED
else:
    # These are the same values as the Colorama color codes. Redefining them
    # here allows users to specify that they want color without having to
    # install Colorama, which is only supposed to be necessary in Windows.
    RESET_ALL = "\033[0m"
    BRIGHT = "\033[1m"
    DIM = "\033[2m"
    RED = "\033[31m"
    BLUE = "\033[34m"
    CYAN = "\033[36m"
    MAGENTA = "\033[35m"
    YELLOW = "\033[33m"
    GREEN = "\033[32m"
    RED_BACK = "\033[41m"

# On Windows, colors are only available if Colorama is installed.
_has_colors = not _IS_WINDOWS or colorama is not None

# Prevent breakage of packages that used the old name of the variable.
_use_colors = _has_colors


class _Styles(Protocol):
    reset: str
    bright: str
    level_critical: str
    level_exception: str
    level_error: str
    level_warn: str
    level_info: str
    level_debug: str
    level_notset: str

    timestamp: str
    logger_name: str
    kv_key: str
    kv_value: str


Styles = Union[_Styles, Type[_Styles]]


class _ColorfulStyles:
    reset = RESET_ALL
    bright = BRIGHT

    level_critical = RED
    level_exception = RED
    level_error = RED
    level_warn = YELLOW
    level_info = GREEN
    level_debug = GREEN
    level_notset = RED_BACK

    timestamp = DIM
    logger_name = BLUE
    kv_key = CYAN
    kv_value = MAGENTA


class _PlainStyles:
    reset = ""
    bright = ""

    level_critical = ""
    level_exception = ""
    level_error = ""
    level_warn = ""
    level_info = ""
    level_debug = ""
    level_notset = ""

    timestamp = ""
    logger_name = ""
    kv_key = ""
    kv_value = ""


class ColumnFormatter(Protocol):
    """
    :class:`~typing.Protocol` for column formatters.

    See `KeyValueColumnFormatter` and `LogLevelColumnFormatter` for examples.

    .. versionadded:: 23.3.0
    """

    def __call__(self, key: str, value: object) -> str:
        """
        Format *value* for *key*.

        This method is responsible for formatting, *key*, the ``=``, and the
        *value*. That means that it can use any string instead of the ``=`` and
        it can leave out both the *key* or the *value*.

        If it returns an empty string, the column is omitted completely.
        """


@dataclass
class Column:
    """
    A column defines the way a key-value pair is formatted, and, by it's
    position to the *columns* argument of `ConsoleRenderer`, the order in which
    it is rendered.

    Args:
        key:
            The key for which this column is responsible. Leave empty to define
            it as the default formatter.

        formatter: The formatter for columns with *key*.

    .. versionadded:: 23.3.0
    """

    key: str
    formatter: ColumnFormatter


@dataclass
class KeyValueColumnFormatter:
    """
    Format a key-value pair.

    Args:
        key_style: The style to apply to the key. If None, the key is omitted.

        value_style: The style to apply to the value.

        reset_style: The style to apply whenever a style is no longer needed.

        value_repr:
            A callable that returns the string representation of the value.

        width: The width to pad the value to. If 0, no padding is done.

        prefix:
            A string to prepend to the formatted key-value pair. May contain
            styles.

        postfix:
            A string to append to the formatted key-value pair. May contain
            styles.

    .. versionadded:: 23.3.0
    """

    key_style: str | None
    value_style: str
    reset_style: str
    value_repr: Callable[[object], str]
    width: int = 0
    prefix: str = ""
    postfix: str = ""

    def __call__(self, key: str, value: object) -> str:
        sio = StringIO()

        if self.prefix:
            sio.write(self.prefix)
            sio.write(self.reset_style)

        if self.key_style is not None:
            sio.write(self.key_style)
            sio.write(key)
            sio.write(self.reset_style)
            sio.write("=")

        sio.write(self.value_style)
        sio.write(_pad(self.value_repr(value), self.width))
        sio.write(self.reset_style)

        if self.postfix:
            sio.write(self.postfix)
            sio.write(self.reset_style)

        return sio.getvalue()


class LogLevelColumnFormatter:
    """
    Format a log level according to *level_styles*.

    The width is padded to the longest level name (if *level_styles* is passed
    -- otherwise there's no way to know the lengths of all levels).

    Args:
        level_styles:
            A dictionary of level names to styles that are applied to it. If
            None, the level is formatted as a plain ``[level]``.

        reset_style:
            What to use to reset the style after the level name. Ignored if
            if *level_styles* is None.

        width:
            The width to pad the level to. If 0, no padding is done.

    .. versionadded:: 23.3.0
    .. versionadded:: 24.2.0 *width*
    """

    level_styles: dict[str, str] | None
    reset_style: str
    width: int

    def __init__(
        self,
        level_styles: dict[str, str],
        reset_style: str,
        width: int | None = None,
    ) -> None:
        self.level_styles = level_styles
        if level_styles:
            self.width = (
                0
                if width == 0
                else len(max(self.level_styles.keys(), key=lambda e: len(e)))
            )
            self.reset_style = reset_style
        else:
            self.width = 0
            self.reset_style = ""

    def __call__(self, key: str, value: object) -> str:
        level = cast(str, value)
        style = (
            ""
            if self.level_styles is None
            else self.level_styles.get(level, "")
        )

        return f"[{style}{_pad(level, self.width)}{self.reset_style}]"


_NOTHING = object()


def plain_traceback(sio: TextIO, exc_info: ExcInfo) -> None:
    """
    "Pretty"-print *exc_info* to *sio* using our own plain formatter.

    To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

    Used by default if neither Rich nor *better-exceptions* are present.

    .. versionadded:: 21.2.0
    """
    sio.write("\n" + _format_exception(exc_info))


@dataclass
class RichTracebackFormatter:
    """
    A Rich traceback renderer with the given options.

    Pass an instance as `ConsoleRenderer`'s ``exception_formatter`` argument.

    See :class:`rich.traceback.Traceback` for details on the arguments.

    If a *width* of -1 is passed, the terminal width is used. If the width
    can't be determined, fall back to 80.

    .. versionadded:: 23.2.0
    """

    color_system: Literal[
        "auto", "standard", "256", "truecolor", "windows"
    ] = "truecolor"
    show_locals: bool = True
    max_frames: int = 100
    theme: str | None = None
    word_wrap: bool = False
    extra_lines: int = 3
    width: int = 100
    indent_guides: bool = True
    locals_max_length: int = 10
    locals_max_string: int = 80
    locals_hide_dunder: bool = True
    locals_hide_sunder: bool = False
    suppress: Sequence[str | ModuleType] = ()

    def __call__(self, sio: TextIO, exc_info: ExcInfo) -> None:
        if self.width == -1:
            self.width, _ = shutil.get_terminal_size((80, 0))

        sio.write("\n")

        Console(
            file=sio, color_system=self.color_system, width=self.width
        ).print(
            Traceback.from_exception(
                *exc_info,
                show_locals=self.show_locals,
                max_frames=self.max_frames,
                theme=self.theme,
                word_wrap=self.word_wrap,
                extra_lines=self.extra_lines,
                width=self.width,
                indent_guides=self.indent_guides,
                locals_max_length=self.locals_max_length,
                locals_max_string=self.locals_max_string,
                locals_hide_dunder=self.locals_hide_dunder,
                locals_hide_sunder=self.locals_hide_sunder,
                suppress=self.suppress,
            )
        )


rich_traceback = RichTracebackFormatter()
"""
Pretty-print *exc_info* to *sio* using the Rich package.

To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

This is a `RichTracebackFormatter` with default arguments and used by default
if Rich is installed.

.. versionadded:: 21.2.0
"""


def better_traceback(sio: TextIO, exc_info: ExcInfo) -> None:
    """
    Pretty-print *exc_info* to *sio* using the *better-exceptions* package.

    To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

    Used by default if *better-exceptions* is installed and Rich is absent.

    .. versionadded:: 21.2.0
    """
    sio.write("\n" + "".join(better_exceptions.format_exception(*exc_info)))


if rich is not None:
    default_exception_formatter = rich_traceback
elif better_exceptions is not None:
    default_exception_formatter = better_traceback
else:
    default_exception_formatter = plain_traceback


class ConsoleRenderer:
    r"""
    Render ``event_dict`` nicely aligned, possibly in colors, and ordered.

    If ``event_dict`` contains a true-ish ``exc_info`` key, it will be rendered
    *after* the log line. If Rich_ or better-exceptions_ are present, in colors
    and with extra context.

    Args:
        columns:
            A list of `Column` objects defining both the order and format of
            the key-value pairs in the output. If passed, most other arguments
            become meaningless.

            **Must** contain a column with ``key=''`` that defines the default
            formatter.

            .. seealso:: `columns-config`

        pad_event:
            Pad the event to this many characters. Ignored if *columns* are
            passed.

        colors:
            Use colors for a nicer output. `True` by default. On Windows only
            if Colorama_ is installed. Ignored if *columns* are passed.

        force_colors:
            Force colors even for non-tty destinations. Use this option if your
            logs are stored in a file that is meant to be streamed to the
            console. Only meaningful on Windows. Ignored if *columns* are
            passed.

        repr_native_str:
            When `True`, `repr` is also applied to ``str``\ s. The ``event``
            key is *never* `repr` -ed. Ignored if *columns* are passed.

        level_styles:
            When present, use these styles for colors. This must be a dict from
            level names (strings) to Colorama styles. The default can be
            obtained by calling `ConsoleRenderer.get_default_level_styles`.
            Ignored when *columns* are passed.

        exception_formatter:
            A callable to render ``exc_infos``. If Rich_ or better-exceptions_
            are installed, they are used for pretty-printing by default (rich_
            taking precedence). You can also manually set it to
            `plain_traceback`, `better_traceback`, an instance of
            `RichTracebackFormatter` like `rich_traceback`, or implement your
            own.

        sort_keys:
            Whether to sort keys when formatting. `True` by default. Ignored if
            *columns* are passed.

        event_key:
            The key to look for the main log message. Needed when you rename it
            e.g. using `structlog.processors.EventRenamer`. Ignored if
            *columns* are passed.

        timestamp_key:
            The key to look for timestamp of the log message. Needed when you
            rename it e.g. using `structlog.processors.EventRenamer`. Ignored
            if *columns* are passed.

        pad_level:
            Whether to pad log level with blanks to the longest amongst all
            level label.

    Requires the Colorama_ package if *colors* is `True` **on Windows**.

    Raises:
        ValueError: If there's not exactly one default column formatter.

    .. _Colorama: https://pypi.org/project/colorama/
    .. _better-exceptions: https://pypi.org/project/better-exceptions/
    .. _Rich: https://pypi.org/project/rich/

    .. versionadded:: 16.0.0
    .. versionadded:: 16.1.0 *colors*
    .. versionadded:: 17.1.0 *repr_native_str*
    .. versionadded:: 18.1.0 *force_colors*
    .. versionadded:: 18.1.0 *level_styles*
    .. versionchanged:: 19.2.0
       Colorama now initializes lazily to avoid unwanted initializations as
       ``ConsoleRenderer`` is used by default.
    .. versionchanged:: 19.2.0 Can be pickled now.
    .. versionchanged:: 20.1.0
       Colorama does not initialize lazily on Windows anymore because it breaks
       rendering.
    .. versionchanged:: 21.1.0
       It is additionally possible to set the logger name using the
       ``logger_name`` key in the ``event_dict``.
    .. versionadded:: 21.2.0 *exception_formatter*
    .. versionchanged:: 21.2.0
       `ConsoleRenderer` now handles the ``exc_info`` event dict key itself. Do
       **not** use the `structlog.processors.format_exc_info` processor
       together with `ConsoleRenderer` anymore! It will keep working, but you
       can't have customize exception formatting and a warning will be raised
       if you ask for it.
    .. versionchanged:: 21.2.0
       The colors keyword now defaults to True on non-Windows systems, and
       either True or False in Windows depending on whether Colorama is
       installed.
    .. versionadded:: 21.3.0 *sort_keys*
    .. versionadded:: 22.1.0 *event_key*
    .. versionadded:: 23.2.0 *timestamp_key*
    .. versionadded:: 23.3.0 *columns*
    .. versionadded:: 24.2.0 *pad_level*
    """

    def __init__(  # noqa: PLR0912, PLR0915
        self,
        pad_event: int = _EVENT_WIDTH,
        colors: bool = _has_colors,
        force_colors: bool = False,
        repr_native_str: bool = False,
        level_styles: Styles | None = None,
        exception_formatter: ExceptionRenderer = default_exception_formatter,
        sort_keys: bool = True,
        event_key: str = "event",
        timestamp_key: str = "timestamp",
        columns: list[Column] | None = None,
        pad_level: bool = True,
    ):
        self._exception_formatter = exception_formatter
        self._sort_keys = sort_keys

        if columns is not None:
            to_warn = []

            def add_meaningless_arg(arg: str) -> None:
                to_warn.append(
                    f"The `{arg}` argument is ignored when passing `columns`.",
                )

            if pad_event != _EVENT_WIDTH:
                add_meaningless_arg("pad_event")

            if colors != _has_colors:
                add_meaningless_arg("colors")

            if force_colors is not False:
                add_meaningless_arg("force_colors")

            if repr_native_str is not False:
                add_meaningless_arg("repr_native_str")

            if level_styles is not None:
                add_meaningless_arg("level_styles")

            if event_key != "event":
                add_meaningless_arg("event_key")

            if timestamp_key != "timestamp":
                add_meaningless_arg("timestamp_key")

            for w in to_warn:
                warnings.warn(w, stacklevel=2)

            defaults = [col for col in columns if col.key == ""]
            if not defaults:
                raise ValueError(
                    "Must pass a default column formatter (a column with `key=''`)."
                )
            if len(defaults) > 1:
                raise ValueError("Only one default column formatter allowed.")

            self._default_column_formatter = defaults[0].formatter
            self._columns = [col for col in columns if col.key]

            return

        # Create default columns configuration.
        styles: Styles
        if colors:
            if _IS_WINDOWS:  # pragma: no cover
                # On Windows, we can't do colorful output without colorama.
                if colorama is None:
                    classname = self.__class__.__name__
                    raise SystemError(
                        _MISSING.format(
                            who=classname + " with `colors=True`",
                            package="colorama",
                        )
                    )
                # Colorama must be init'd on Windows, but must NOT be
                # init'd on other OSes, because it can break colors.
                if force_colors:
                    colorama.deinit()
                    colorama.init(strip=False)
                else:
                    colorama.init()

            styles = _ColorfulStyles
        else:
            styles = _PlainStyles

        self._styles = styles

        level_to_color = (
            self.get_default_level_styles(colors)
            if level_styles is None
            else level_styles
        )

        for key in level_to_color:
            level_to_color[key] += styles.bright
        self._longest_level = len(
            max(level_to_color.keys(), key=lambda e: len(e))
        )

        self._repr_native_str = repr_native_str

        self._default_column_formatter = KeyValueColumnFormatter(
            styles.kv_key,
            styles.kv_value,
            styles.reset,
            value_repr=self._repr,
            width=0,
        )

        logger_name_formatter = KeyValueColumnFormatter(
            key_style=None,
            value_style=styles.bright + styles.logger_name,
            reset_style=styles.reset,
            value_repr=str,
            prefix="[",
            postfix="]",
        )

        level_width = 0 if not pad_level else None

        self._columns = [
            Column(
                timestamp_key,
                KeyValueColumnFormatter(
                    key_style=None,
                    value_style=styles.timestamp,
                    reset_style=styles.reset,
                    value_repr=str,
                ),
            ),
            Column(
                "level",
                LogLevelColumnFormatter(
                    level_to_color, reset_style=styles.reset, width=level_width
                ),
            ),
            Column(
                event_key,
                KeyValueColumnFormatter(
                    key_style=None,
                    value_style=styles.bright,
                    reset_style=styles.reset,
                    value_repr=str,
                    width=pad_event,
                ),
            ),
            Column("logger", logger_name_formatter),
            Column("logger_name", logger_name_formatter),
        ]

    def _repr(self, val: Any) -> str:
        """
        Determine representation of *val* depending on its type &
        self._repr_native_str.
        """
        if self._repr_native_str is True:
            return repr(val)

        if isinstance(val, str):
            return val

        return repr(val)

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> str:
        stack = event_dict.pop("stack", None)
        exc = event_dict.pop("exception", None)
        exc_info = event_dict.pop("exc_info", None)

        kvs = [
            col.formatter(col.key, val)
            for col in self._columns
            if (val := event_dict.pop(col.key, _NOTHING)) is not _NOTHING
        ] + [
            self._default_column_formatter(key, event_dict[key])
            for key in (sorted(event_dict) if self._sort_keys else event_dict)
        ]

        sio = StringIO()
        sio.write((" ".join(kv for kv in kvs if kv)).rstrip(" "))

        if stack is not None:
            sio.write("\n" + stack)
            if exc_info or exc is not None:
                sio.write("\n\n" + "=" * 79 + "\n")

        if exc_info:
            exc_info = _figure_out_exc_info(exc_info)

            if exc_info != (None, None, None):
                self._exception_formatter(sio, exc_info)
        elif exc is not None:
            if self._exception_formatter is not plain_traceback:
                warnings.warn(
                    "Remove `format_exc_info` from your processor chain "
                    "if you want pretty exceptions.",
                    stacklevel=2,
                )

            sio.write("\n" + exc)

        return sio.getvalue()

    @staticmethod
    def get_default_level_styles(colors: bool = True) -> Any:
        """
        Get the default styles for log levels

        This is intended to be used with `ConsoleRenderer`'s ``level_styles``
        parameter.  For example, if you are adding custom levels in your
        home-grown :func:`~structlog.stdlib.add_log_level` you could do::

            my_styles = ConsoleRenderer.get_default_level_styles()
            my_styles["EVERYTHING_IS_ON_FIRE"] = my_styles["critical"] renderer
            = ConsoleRenderer(level_styles=my_styles)

        Args:
            colors:
                Whether to use colorful styles. This must match the *colors*
                parameter to `ConsoleRenderer`. Default: `True`.
        """
        styles: Styles
        styles = _ColorfulStyles if colors else _PlainStyles
        return {
            "critical": styles.level_critical,
            "exception": styles.level_exception,
            "error": styles.level_error,
            "warn": styles.level_warn,
            "warning": styles.level_warn,
            "info": styles.level_info,
            "debug": styles.level_debug,
            "notset": styles.level_notset,
        }


_SENTINEL = object()


def set_exc_info(
    logger: WrappedLogger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Set ``event_dict["exc_info"] = True`` if *method_name* is ``"exception"``.

    Do nothing if the name is different or ``exc_info`` is already set.
    """
    if (
        method_name != "exception"
        or event_dict.get("exc_info", _SENTINEL) is not _SENTINEL
    ):
        return event_dict

    event_dict["exc_info"] = True

    return event_dict



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .type_checked_property import TypeCheckedProperty

__all__ = ["TypeCheckedProperty"]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.assess_extensions import cs_str
from contrast.agent import scope
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import patch_manager
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.utils.namespace import Namespace


MAX_SUBCLASS_SEARCH_DEPTH = 32


class module(Namespace):
    patched_classes = []


def _find_subclasses(type, *, _depth=MAX_SUBCLASS_SEARCH_DEPTH):
    if _depth == 0:
        return
    for subclass in type.__subclasses__():
        yield subclass
        yield from _find_subclasses(subclass, _depth=_depth - 1)


def check_or_enter_scope(func):
    def wrapper(*args, **kwargs):
        if scope.in_contrast_or_propagation_scope():
            return
        with scope.contrast_scope(), scope.propagation_scope():
            func(*args, **kwargs)
            return

    return wrapper


@check_or_enter_scope
@fail_quietly("Failed to propagate string subclass __new__")
def propagate_str_cast(result, args, kwargs):
    string_propagation.propagate_unicode_cast(result, args[0], result, args[1:], kwargs)


@check_or_enter_scope
@fail_quietly("Failed to propagate bytes subclass __new__")
def propagate_bytes_cast(result, args, kwargs):
    string_propagation.propagate_bytes_cast(result, args[0], result, args[1:], kwargs)


@check_or_enter_scope
@fail_quietly("Failed to propagate bytearray subclass __init__")
def propagate_bytearray_init(instance, args, kwargs):
    """
    To make use of our generalized bytearray cast propagation machinery, we need to call
    `propagate_bytearray_cast` as though this were a __new__ patch. The calling
    semantics of __new__ differ from __init__, so we shuffle some things around
    accordingly.
    """
    target = instance
    result = instance
    self_obj = None
    string_propagation.propagate_bytearray_cast(target, result, self_obj, args, kwargs)


def build_cast_patch(orig_func, patch_policy, propagation_func):
    del patch_policy

    def __new__(*args, **kwargs):
        result = orig_func(*args, **kwargs)
        propagation_func(result, args, kwargs)
        return result

    # NOTE: builtin methods can't be wrapped
    return __new__


def build_bytearray_init_patch(orig_func, patch_policy):
    del patch_policy

    def __init__(self, *args, **kwargs) -> None:
        result = orig_func(self, *args, **kwargs)
        propagate_bytearray_init(self, args, kwargs)
        return result

    return __init__


def _is_immutable(cls) -> bool:
    """
    We currently don't have machinery for patching __init__ or __new__ on immutable
    types. This appears to be sufficiently rare and is not currently an issue.
    """
    flags = getattr(cls, "__flags__", None)
    if flags is None:
        return False

    if sys.version_info >= (3, 10):
        return bool(flags & cs_str.Py_TPFLAGS_IMMUTABLETYPE)
    return not (flags & cs_str.Py_TPFLAGS_HEAPTYPE)


@fail_quietly("Failed to apply patches for strtype subclasses")
def register_patches():
    """
    Applies patches to known subclasses of str, bytes, and bytearray

    Problem statement: the cast/new hook for subclasses of strtypes does not
    always work. This seems to be an issue with timing: subclasses that
    are defined *after* our extension hooks are applied do not appear to have
    a problem.

    However, even with the runner, our instrumentation is not necessarily going
    to be early enough to affect *all* subclasses that are defined.

    The solution is to use __subclasses__() to find any subclasses of strtypes
    that may already be defined at this point in time. We then explicitly apply
    propagation patches to the __new__ (or __init__) methods of these classes.
    """
    for cls in _find_subclasses(str):
        if _is_immutable(cls):
            continue
        build_and_apply_patch(
            cls,
            "__new__",
            build_cast_patch,
            builder_args=(propagate_str_cast,),
        )
        module.patched_classes.append(cls)

    for cls in _find_subclasses(bytes):
        if _is_immutable(cls):
            continue
        build_and_apply_patch(
            cls,
            "__new__",
            build_cast_patch,
            builder_args=(propagate_bytes_cast,),
        )
        module.patched_classes.append(cls)

    for cls in _find_subclasses(bytearray):
        if _is_immutable(cls):
            continue
        build_and_apply_patch(
            cls,
            "__init__",
            build_bytearray_init_patch,
        )
        module.patched_classes.append(cls)


def reverse_patches():
    for cls in module.patched_classes:
        patch_manager.reverse_patches_by_owner(cls)

    module.patched_classes = []



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import platform
import sys

from datetime import datetime, timezone
from typing import Any

import contrast
from contrast import __version__
from contrast.agent.exceptions import _Exception
from contrast.agent.validator import Validator
from contrast.agent.metrics import MetricsDict
from contrast_rewriter import is_rewriter_enabled
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


class TelemetryEvent:
    def __init__(self, instance_id: str):
        self.timestamp = datetime.now(timezone.utc).astimezone().isoformat()
        self.telemetry_instance_id = instance_id
        self.tags = MetricsDict(str)

    @property
    def path(self) -> str:
        return ""

    def add_tags(self) -> None:
        from contrast.agent.agent_state import (
            detected_framework,
            free_threading_available,
            free_threading_enabled,
            jit_available,
            jit_enabled,
        )

        telemetry = contrast.TELEMETRY
        settings = telemetry.settings

        self.tags["is_public_build"] = str(telemetry.is_public_build).lower()
        self.tags["agent_version"] = __version__
        self.tags["python_version"] = sys.version
        self.tags["python_version_simple"] = platform.python_version()
        self.tags["os_type"] = platform.system()
        self.tags["os_arch"] = platform.machine()
        self.tags["os_version"] = platform.version()
        self.tags["app_framework_version"] = str(detected_framework())
        self.tags["server_framework_version"] = str(settings.server)
        self.tags["teamserver"] = settings.config.teamserver_type
        self.tags["rewriter_enabled"] = str(is_rewriter_enabled()).lower()
        self.tags["free_threading_available"] = str(free_threading_available()).lower()
        self.tags["free_threading_enabled"] = str(free_threading_enabled()).lower()
        self.tags["jit_available"] = str(jit_available()).lower()
        self.tags["jit_enabled"] = str(jit_enabled()).lower()

    def to_json(self) -> dict[str, Any]:
        return dict(
            timestamp=self.timestamp,
            instance=self.telemetry_instance_id,
            tags=self.tags,
        )


class MetricsTelemetryEvent(TelemetryEvent):
    def __init__(self, instance_id: str) -> None:
        super().__init__(instance_id)
        self.fields = MetricsDict(int)
        self.fields["_filler"] = 0

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.to_json()}"

    def to_json(self) -> dict[str, Any]:
        return {
            **super().to_json(),
            **dict(fields=self.fields),
        }


class StartupMetricsTelemetryEvent(MetricsTelemetryEvent):
    def __init__(self, instance_id: str) -> None:
        super().__init__(instance_id)

        try:
            self.add_tags()
        except Exception as ex:
            logger.debug("Telemetry failed to create StartupMetrics msg: %s", ex)
            logger.debug("Disabling telemetry")
            contrast.TELEMETRY.enabled = False

    @property
    def path(self) -> str:
        return "/metrics/startup"


class ErrorTelemetryEvent(TelemetryEvent, Validator):
    VALIDATIONS = dict(
        telemetry_instance_id=dict(required=True, range=(12, 64)),
        # tags field is required to be reported but can be empty
        tags=dict(required=True, range=(0, 512)),
        exceptions=dict(required=True, range=(1, 512)),
        # some agents use this for named logger
        logger=dict(required=False, range=(0, 128)),
        # logged message, not exception message
        message=dict(required=False, range=(0, 256)),
        occurrences=dict(required=False, default=0),
    )

    def __init__(
        self, instance_id: str, error: Exception, logger_="", message="", skip_frames=0
    ) -> None:
        super().__init__(instance_id)

        # We may not use this field, it's meant to represent a named logger
        # to easily filter a specific feature
        self.logger = logger_
        # message that was logged, not the exception message
        self.message = message
        self.occurrences = 1
        # +1 to skip the current ErrorTelemetryEvent frame
        self.exceptions = [_Exception(error, skip_frames + 1)]

        try:
            self.add_tags()
        except Exception as ex:
            logger.debug("Adding tags failed: %s", ex)

        self.validate()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.to_json()}"

    @property
    def path(self) -> str:
        return "/exceptions/error"

    def to_json(self) -> dict[str, Any]:
        return {
            **super().to_json(),
            **dict(
                occurrences=self.occurrences,
                logger=self.logger,
                message=self.message,
                exceptions=[ex.to_json() for ex in self.exceptions],
            ),
        }



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.utils.configuration_utils import (
    parse_disabled_rules,
    parse_event_detail_option,
    parse_stacktraces_options,
    str_to_bool,
)


class Assess(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="assess.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="assess.enable_scan_response",
                default_value=True,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="assess.sampling.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="assess.sampling.baseline",
                default_value=5,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.sampling.request_frequency",
                default_value=10,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.sampling.window_ms",
                default_value=180_000,
                type_cast=int,
            ),
            ConfigOption(canonical_name="assess.tags", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="assess.rules.disabled_rules",
                default_value=[],
                type_cast=parse_disabled_rules,
            ),
            ConfigOption(
                canonical_name="assess.stacktraces",
                default_value="ALL",
                type_cast=parse_stacktraces_options,
            ),
            ConfigOption(
                canonical_name="assess.max_context_source_events",
                default_value=100,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.max_propagation_events",
                default_value=2000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.time_limit_threshold",
                default_value=300000,  # 5 minutes in ms
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.max_rule_reported",
                default_value=100,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="assess.event_detail",
                default_value="minimal",
                type_cast=parse_event_detail_option,
            ),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import base64

from contrast.utils.string_utils import ensure_string

B64_NONE_STRING = base64.b64encode(str(None).encode("utf-8")).decode("ascii")


def base64_encode(to_encode) -> str:
    """
    Allows you to encode string to Base64
    """
    if to_encode is None:
        return B64_NONE_STRING
    if isinstance(to_encode, str):
        input_object = to_encode.encode("utf-8")
    else:
        input_object = to_encode

    output_object = base64.b64encode(input_object)

    return output_object.decode("ascii")


def base64_decode(to_decode):
    """
    Allows you to decode Base64 to str
    """
    output_object = base64.urlsafe_b64decode(ensure_string(to_decode))
    return ensure_string(output_object)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import TYPE_CHECKING, Callable
import contrast
from contrast.agent import scope
from contrast.agent.settings import Settings

if TYPE_CHECKING:
    from contrast.agent.policy.patch_location_policy import PatchLocationPolicy


def apply_rule(
    patch_policy: "PatchLocationPolicy",
    orig_func: Callable,
    args: tuple,
    kwargs: dict[str, object],
):
    context = contrast.REQUEST_CONTEXT.get()

    if (
        context is not None
        and context.protect_enabled
        and not scope.in_contrast_scope()
    ):
        for node in patch_policy.trigger_nodes:
            rule_name = node.rule.name
            rule = Settings().protect_rules.get(rule_name)
            if not rule or not rule.enabled:
                continue

            for source in node.get_protect_sources(args, kwargs):
                rule.protect(patch_policy, source, args, kwargs)

    return orig_func(*args, **kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import sys
from types import CodeType
from typing import Any
from contrast.utils.assess.tracking_util import SUPPORTED_TYPES
from contrast.utils.decorators import fail_quietly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

# sys.monitoring only allows "tools" with IDs 0-5, inclusive. The following are defined:
#
# sys.monitoring.DEBUGGER_ID = 0
# sys.monitoring.COVERAGE_ID = 1
# sys.monitoring.PROFILER_ID = 2
# sys.monitoring.OPTIMIZER_ID = 5
#
# However, any value from 0-5 may be used. To reduce the probability of conflict, and
# since we don't really fall into any of the above tool types, we use a different ID.
# See https://docs.python.org/3/library/sys.monitoring.html
MAX_TOOL_ID = 5
CONTRAST_TOOL_ID = 4
CONTRAST_TOOL_NAME = "contrast"
VALID_TOOL_IDS = range(MAX_TOOL_ID + 1)


@fail_quietly()
def enable() -> None:
    """
    Enable sys.monitoring events if possible. We currently use this as a workaround for
    string propagation loss that results from bytecode specialization introduced in
    python 3.12. Enabling sys.monitoring prevents bytecode specialization from occurring
    for the monitored instructions. See the related pytest fixture for more detail.

    This is a fairly blunt solution to this problem, and in the future it may not be
    necessary.

    Returns True if sys.monitoring is enabled successfully and False otherwise.
    """
    if sys.version_info[:2] < (3, 12):
        logger.debug("Pre-3.12 environment detected. Will not enable sys.monitoring.")
        return

    if is_enabled():
        logger.debug("sys.monitoring is already enabled.")
        return

    from sys import monitoring

    logger.debug("summarizing sys.monitoring tools already in use")
    for tool_id in VALID_TOOL_IDS:
        logger.debug(
            "sys.monitoring tool ID %s: %s", tool_id, monitoring.get_tool(tool_id)
        )

    if monitoring.get_tool(CONTRAST_TOOL_ID) is not None:
        logger.debug(
            "Cannot enable sys.monitoring - tool ID %s is already registered",
            CONTRAST_TOOL_ID,
        )
        return

    monitoring.use_tool_id(CONTRAST_TOOL_ID, CONTRAST_TOOL_NAME)
    monitoring.set_events(CONTRAST_TOOL_ID, monitoring.events.CALL)
    monitoring.register_callback(
        CONTRAST_TOOL_ID, monitoring.events.CALL, _strlike_call_specialization_disabler
    )

    logger.debug("Enabled contrast sys.monitoring tool")


def _strlike_call_specialization_disabler(
    code: CodeType, instruction_offset: int, callable: object, arg0: object
) -> Any:
    # We need to check SUPPORTED_TYPES is not None because this callback will
    # still be installed during interpreter shutdown, when global variables may be
    # set to None.
    if SUPPORTED_TYPES is not None and callable not in SUPPORTED_TYPES:
        return sys.monitoring.DISABLE


@fail_quietly(return_value=False)
def is_enabled() -> bool:
    if sys.version_info[:2] < (3, 12):
        return False
    from sys import monitoring

    return bool(
        monitoring.get_tool(CONTRAST_TOOL_ID) == CONTRAST_TOOL_NAME
        and monitoring.get_events(CONTRAST_TOOL_ID) & monitoring.events.CALL
    )


@fail_quietly()
def disable():
    if sys.version_info[:2] < (3, 12):
        return
    from sys import monitoring

    monitoring.set_events(CONTRAST_TOOL_ID, monitoring.events.NO_EVENTS)
    monitoring.free_tool_id(CONTRAST_TOOL_ID)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging
from contrast_fireball import LibraryObservation

logger = logging.getLogger("contrast")


class LibraryUsage(BaseTsAppMessage):
    def __init__(self, observed_libraries: list[LibraryObservation]):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()

        self.base_url = f"{self.settings.api_url}/agents/v1.0/"
        self.body = {
            "observations": [
                {"id": lib.library_hash, "names": lib.names}
                for lib in observed_libraries
                if lib.library_hash and lib.names
            ]
        }

    @property
    def name(self):
        return "library-usage"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "library-usage",
            ]
        )

    @property
    def request_method(self):
        return requests.post

    @fail_loudly("Failed to process LibraryUsage response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
from contrast.agent import scope
from contrast.agent.settings import Settings
from contrast.reporting.teamserver_messages import Heartbeat
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import now_ms, sleep

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")
HEARTBEAT_THREAD_NAME = "ContrastHeartbeat"


class HeartbeatThread(threading.Thread):
    def __init__(self, reporting_client):
        self.stopped = False
        self.heartbeat_interval_ms = Settings().config.heartbeat_poll_interval
        # Agent should not ping too frequently
        self.heartbeat_interval_ms = max(10000, self.heartbeat_interval_ms)

        super().__init__()
        # A thread must have had __init__ called, but not start, to set daemon
        self.daemon = True
        self.name = HEARTBEAT_THREAD_NAME
        self.reporting_client = reporting_client

    def start(self):
        self.stopped = False
        super().start()

    @property
    def settings_interval_sec(self):
        return self.heartbeat_interval_ms / 1000

    def run(self):
        # Ensure the heartbeat thread runs in scope because it is initialized
        # before our thread.start patch is applied.
        with scope.contrast_scope():
            logger.debug("Establishing heartbeat")

            while not self.stopped and Settings().is_agent_config_enabled():
                self.send_heartbeat()
                sleep(self.settings_interval_sec)

    @fail_loudly("Error sending a heartbeat message")
    def send_heartbeat(self):
        settings = Settings()
        if settings.config is None:
            return

        # only send after we have not updated after the last interval time set in the config
        if (
            settings.last_app_update_time_ms is None
            or (now_ms() - settings.last_app_update_time_ms)
            < self.heartbeat_interval_ms
        ):
            return

        msg = Heartbeat()
        response = self.reporting_client.send_message(msg)
        msg.process_response(response, self.reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from dataclasses import asdict
from typing import Any
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class _Traces(BaseTsAppMessage):
    """This class should only be instantiated by Preflight's response handler"""

    def __init__(self, finding, reportable_request):
        super().__init__()

        self.extra_headers["Report-Hash"] = str(finding.hash)

        self.body = self._build_body(finding, reportable_request)

    @property
    def name(self) -> str:
        return "traces"

    @property
    def path(self) -> str:
        return "traces"

    @property
    def request_method(self):
        return requests.put

    @fail_loudly("Failed to process Traces response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)

    def _build_body(self, finding, reportable_request):
        base_body = {
            "created": finding.created,
            "events": [as_camel_dict(event) for event in finding.events],
            "properties": finding.properties,
            "routes": [as_camel_dict(route) for route in finding.routes],
            "ruleId": finding.rule_id,
            "tags": finding.tags,
            "version": finding.version,
        }
        if session_id := self.settings.config.session_id:
            base_body["session_id"] = session_id
        if reportable_request is not None:
            base_body["request"] = reportable_request

        return base_body


def as_camel_dict(data):
    return asdict(data, dict_factory=lower_camel_case_keyed_dict_factory)


def lower_camel_case_keyed_dict_factory(kvs: list[tuple[str, Any]]) -> dict[str, Any]:
    return {lower_snake_to_camel(k): v for k, v in kvs if v is not None}


def lower_snake_to_camel(snake_str: str) -> str:
    return "".join(
        word if i == 0 else word.capitalize()
        for i, word in enumerate(snake_str.split("_"))
    )



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Extracted log level data used by both stdlib and native log level filters.
"""

from __future__ import annotations

import logging

from .typing import EventDict


# Adapted from the stdlib
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

NAME_TO_LEVEL = {
    "critical": CRITICAL,
    "exception": ERROR,
    "error": ERROR,
    "warn": WARNING,
    "warning": WARNING,
    "info": INFO,
    "debug": DEBUG,
    "notset": NOTSET,
}

LEVEL_TO_NAME = {
    v: k
    for k, v in NAME_TO_LEVEL.items()
    if k not in ("warn", "exception", "notset")
}

# Keep around for backwards-compatability in case someone imported them.
_LEVEL_TO_NAME = LEVEL_TO_NAME
_NAME_TO_LEVEL = NAME_TO_LEVEL


def map_method_name(method_name: str) -> str:
    # warn is just a deprecated alias in the stdlib.
    if method_name == "warn":
        return "warning"

    # Calling exception("") is the same as error("", exc_info=True)
    if method_name == "exception":
        return "error"

    return method_name


def add_log_level(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Add the log level to the event dict under the ``level`` key.

    Since that's just the log method name, this processor works with non-stdlib
    logging as well. Therefore it's importable both from `structlog.processors`
    as well as from `structlog.stdlib`.

    .. versionadded:: 15.0.0
    .. versionchanged:: 20.2.0
       Importable from `structlog.processors` (additionally to
       `structlog.stdlib`).
    .. versionchanged:: 24.1.0
       Added mapping from "exception" to "error"
    """

    event_dict["level"] = map_method_name(method_name)

    return event_dict



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Propagator used for flask.helpers.safe_join

Tests for this propagator are found in
tests/agent/assess/sanitizers/test_flask_sanitizers.py
"""

from functools import cached_property
import posixpath
from os import path
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators import JoinPropagator
from contrast.utils.string_utils import ensure_string


class SafeJoinPropagator(JoinPropagator):
    """
    Sanitizing propagator for flask.helpers.safe_join

    The safe_join function is used to join path components to a base directory in such
    a way that a path traversal attack is not possible. The tag propagation from
    sources to target is the same as the basic str.join propagation. Unlike str.join,
    in the case of safe_join, the separator is not given as a source. Instead, we
    simply set it to posixpath.sep since that is what is used by safe_join under the
    hood.

    The most important difference between str.join and safe_join is the tags that are
    applied. The safe_join propagator adds sanitizing tags since the path components
    passed to safe_join can't contribute to a path traversal attack. However, safe_join
    _assumes_ that the base directory is trusted. We can't make that same assumption
    since it actually can come from anywhere, including an untrusted source. Since the
    base directory isn't validated by safe_join in any way, we do not apply safe tags
    to the corresponding range in the target.
    """

    use_abspath = False

    def __init__(self, node, preshift, target):
        super().__init__(node, preshift, target)
        # safe_join looks like '/'.join(directory, *pathnames)
        # Under the hood, safe_join uses posixpath.sep as the separator, so we use it
        # as the separator here as well. However, the value of the separator doesn't
        # actually matter for the purposes of propagation, as long as len(sep) == 1
        self.separator = posixpath.sep
        self.strings_to_join = self.sources
        self.directory = self.preshift.args[0] or ""

    @cached_property
    def sources(self):
        """
        Use all arguments as sources: safe_join(directory, *pathnames)
        """
        # We want to use BasePropagator.sources, not JoinPropagator.sources
        return super(JoinPropagator, self).sources

    def add_tags(self):
        """
        Add sanitizing tags to entire result excluding the base directory
        In some cases the directory has a relative path `..` so we look for the
        absolute path.
        """
        directory_str = ensure_string(self.directory)  # accounts for PosixPath

        directory = (
            path.abspath(directory_str)
            if self.use_abspath and "/.." in directory_str or "/." in directory_str
            else directory_str
        )

        tag_span = AdjustedSpan(len(directory), len(self.target))
        if self.node.tags:
            self.apply_tags(self.node, self.target, tag_span)

        if self.node.untags:
            self.apply_untags(self.node, self.target, tag_span)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections.abc import Iterable
from types import ModuleType

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    unregister_module_patcher,
    wrap_and_watermark,
    register_module_patcher,
)
from contrast.utils.decorators import fail_quietly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

AIOHTTP_WEB_MODULE = "aiohttp.web"


def build__init__patch(orig_func, patch_policy):
    """
    Patch for aiohttp.web.Application.__init__

    This performs automatic middleware installation for aiohttp.
    """
    del patch_policy

    def __init__patch(wrapped, instance, args, kwargs) -> None:
        del instance

        add_middleware_to_kwargs(kwargs)
        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, __init__patch)


@fail_quietly()
def add_middleware_to_kwargs(kwargs) -> None:
    """
    To actually auto-instrument aiohttp, we need to forcefully add our middleware to
    the `middlewares` kwarg passed in to aiohttp.web.Application.__init__. This kwarg
    expects an iterable of initialized aiohttp-style middlewares.

    This function simply accepts the constructor's original kwargs dict and injects our
    AioHttpMiddleware into the `middlewares` kwarg.
    """
    from contrast.aiohttp.middleware import AioHttpMiddleware
    from contrast.agent.agent_state import automatic_middleware

    contrast_middleware: AioHttpMiddleware
    with automatic_middleware():
        contrast_middleware = AioHttpMiddleware()

    middlewares: Iterable = kwargs.get("middlewares")
    if not middlewares:
        kwargs["middlewares"] = [contrast_middleware]
    elif isinstance(middlewares, list):
        middlewares.append(contrast_middleware)
    else:
        logger.error(
            "Cannot automatically install ContrastMiddleware - "
            "aiohttp's middlewares list had an unexpected type. "
            "See Contrast's documentation for installing the middleware manually."
        )


def patch_aiohttp_autoinstrumentation(aiohttp_web_module: ModuleType) -> None:
    build_and_apply_patch(
        aiohttp_web_module.Application, "__init__", build__init__patch
    )


def register_patches() -> None:
    register_module_patcher(patch_aiohttp_autoinstrumentation, AIOHTTP_WEB_MODULE)


def reverse_patches() -> None:
    unregister_module_patcher(AIOHTTP_WEB_MODULE)
    patch_manager.reverse_class_patches_by_name(AIOHTTP_WEB_MODULE, "Application")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from contrast.utils.decorators import fail_quietly
from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    build_signature,
)
from contrast_fireball import DiscoveredRoute
from pyramid.interfaces import IView, IViewClassifier, IRouteRequest, IRoutesMapper
from zope.interface import Interface

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def get_iface_for_route(registry, pyramid_route):
    return registry.queryUtility(IRouteRequest, name=pyramid_route.name)


def _get_view_func(registry, pyramid_route):
    request_iface = get_iface_for_route(registry, pyramid_route)

    if request_iface is None:
        view_func = None
    else:
        view_func = registry.adapters.lookup(
            (IViewClassifier, request_iface, Interface), IView, name="", default=None
        )
    return view_func


def create_pyramid_routes(registry) -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a pyramid app registry
    """
    routes = set()
    mapper = registry.queryUtility(IRoutesMapper)

    if mapper is None:
        return routes

    for pyramid_route in mapper.get_routes():
        view_func = _get_view_func(registry, pyramid_route)

        if view_func is not None:
            signature = build_signature(pyramid_route.name, view_func)
            path_template = pyramid_route.path
            methods = pyramid_route.predicates or DEFAULT_ROUTE_METHODS
            for method_type in methods:
                routes.add(
                    DiscoveredRoute(
                        verb=method_type,
                        url=path_template,
                        signature=signature,
                        framework="Pyramid",
                    )
                )
        else:
            logger.debug("Unable to add %s to route discovery.", pyramid_route.path)

    return routes


@fail_quietly()
def get_signature_and_path_template(
    request_path, routes_list, registry
) -> tuple[str | None, str | None]:
    if not request_path:
        logger.debug("No path info for pyramid request")
        return None, None

    # Ideally we would like to call get_route but
    # there is no direct relationship between the wsgi
    # request path and the name of the route so we must
    # iterate over all the routes to find it by the path
    # pyramid_route = mapper.get_route(name)

    # TODO: PYT-3826 this is not a robust route matching strategy
    matching_routes = [x for x in routes_list if x.path == request_path]

    if not matching_routes:
        return None, None

    pyramid_route = matching_routes[0]
    if (view_func := _get_view_func(registry, pyramid_route)) is None:
        return None, None
    return build_signature(pyramid_route.name, view_func), pyramid_route.path



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .application_settings import ApplicationSettings
from .ng_server_settings import NGServerSettings
from .protect_rule import ProtectRule
from .server_settings import ServerSettings

__all__ = [
    "ApplicationSettings",
    "NGServerSettings",
    "ProtectRule",
    "ServerSettings",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_response_rule import BaseResponseRule


class BaseBodyOnlyRule(BaseResponseRule):
    @property
    def name(self):
        raise NotImplementedError

    def is_body_violated(self, body, form_tags):
        raise NotImplementedError

    def is_violated(self, _, body, form_tags, meta_tags):
        body_violated, body_properties = self.is_body_violated(body, form_tags)
        return body_violated, body_properties

    def build_properties(self, full_tag, body):
        original_start = body.index(full_tag)
        original_end = original_start + len(full_tag)

        # 50chars before + full tag + 50 chars after
        html = body[max(0, original_start - 50) : min(len(body), original_end + 50)]

        redacted_start = html.index(full_tag)
        redacted_end = redacted_start + len(full_tag)

        return dict(
            html=html,
            start=str(redacted_start),
            end=str(redacted_end),
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Instrumentation for exec/eval.

This was defined outside of policy.json because we need to pass globals/locals from the
frame in which function was originally called
"""

import __future__

import ast
from contextlib import contextmanager
from functools import reduce
import operator
from sys import _getframe as getframe
from types import CodeType, FrameType, ModuleType
from collections.abc import Iterator

import builtins
import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.applies.assess.unsafe_code_execution import (
    apply_rule as apply_unsafe_code_exec_rule,
)
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    get_arg,
    set_arg,
    register_module_patcher,
    unregister_module_patcher,
)
from contrast_rewriter import PropagationRewriter, add_dependency_to_globals


orig_compile = builtins.compile


@scope.contrast_scope()
@fail_quietly("Error applying rule for exec/eval patch")
def apply_rule(rule_applicator, orig_func, result, args, kwargs):
    context = contrast.REQUEST_CONTEXT.get()
    with scope.pop_contrast_scope():
        # need to be in original scope to correctly check if we need to perform analysis
        if not (context and context.propagate_assess):
            return

    rule_applicator("builtins", orig_func.__name__, result, args, kwargs)


def rewrite(mode, code) -> tuple[CodeType, set[ModuleType]]:
    rewriter = PropagationRewriter()
    rewritten_code = rewriter.visit(ast.parse(code, mode=mode))
    return rewritten_code, rewriter.injected_modules


@scope.contrast_scope()
@fail_quietly("Error when applying rewriter to eval/exec/compiled input")
def apply_rewriter(mode, code, flags: int) -> tuple[CodeType, set[ModuleType]]:
    if not isinstance(code, str):
        return code, PropagationRewriter.all_possible_injected_modules

    rewritten_code, needs_modules = rewrite(mode, code)
    return (
        orig_compile(
            rewritten_code,
            filename="<internal>",
            mode=mode,
            flags=flags,
            # inheriting would cause changes in this module
            # to apply to the user code being compiled.
            dont_inherit=1,
        ),
        needs_modules,
    )


_FUTURE_FLAGS = reduce(
    operator.or_,
    [
        getattr(__future__, feature_name).compiler_flag
        for feature_name in __future__.all_feature_names
    ],
    0,
)


def future_flags(code: CodeType) -> int:
    """
    Extract the flags for future features from a code object.

    See https://docs.python.org/3/library/functions.html#compile
    documentation of "flags" parameter for more information.

    Also see "compute_code_flags" in Python's compile.c for
    the implementation of inheritance.
    """
    return code.co_flags & _FUTURE_FLAGS


# This is a context manager to guarantee that the frame is deleted.
# If the frame isn't deleted, we might end up with a reference cycle
# that would cascade into a memory leak, since the frame also holds
# references to the code object and the globals/locals.
# See the Note in https://docs.python.org/3/library/inspect.html#inspect.Traceback
@contextmanager
def calling_frame() -> Iterator[FrameType]:
    """
    Safely yields the frame that called into instrumentation.
    """
    # getframe at depth=2 because
    #   1 = context manager (caller of calling_frame)
    #   2 = patch function (caller of context manager)
    #   3 = user code (caller of patch function)
    frame = getframe(3)
    try:
        yield frame
    finally:
        del frame


def build_exec_eval_patch(orig_func, _, rule_applicator, mode):
    def exec_eval_patch(source, /, globals=None, locals=None, *, closure=None):
        """
        Run exec/eval call with proper context to adjust for current frame

        Code ported from six module
        See https://github.com/benjaminp/six/blob/master/six.py#L694

        Reapplying the context from the 3rd frame (from top of stack) is necessary
        because the globals and locals in that frame are used in the original call to
        exec/eval. The exception to this is if the caller passes custom globals/locals
        to the function.

        If we fail provide this context we will see a number of NameErrors due to things
        not defined in the scope of this function upon calling the original function
        definition.
        """
        result = None

        with calling_frame() as frame:
            if globals is None:
                globals = frame.f_globals
                if locals is None:
                    locals = frame.f_locals
            elif locals is None:
                locals = globals
            co_future_flags = future_flags(frame.f_code)

        try:
            code_to_run, needs_modules = apply_rewriter(mode, source, co_future_flags)
            # Ensure our rewriter patches are in (global) scope for the code about to be executed
            for mod in needs_modules:
                add_dependency_to_globals(globals, mod)
            result = (
                orig_func(code_to_run, globals, locals)
                if closure is None
                else orig_func(code_to_run, globals, locals, closure=closure)
            )
        except Exception:
            result = None
            raise
        finally:
            apply_rule(rule_applicator, orig_func, result, (source,), {})

        return result

    return exec_eval_patch


def build_compile_patch(orig_func, _, rule_applicator):
    def compile_patch(*args, **kwargs):
        if scope.in_contrast_scope():
            return orig_func(*args, **kwargs)

        result = None
        orig_args = args
        orig_kwargs = kwargs.copy()

        try:
            code = get_arg(args, kwargs, 0, kw="source")
            mode = get_arg(args, kwargs, 2, kw="mode", default="exec")
            flags = get_arg(args, kwargs, 3, kw="flags", default=0)
            if flags == 0:
                with calling_frame() as frame:
                    args, kwargs = set_arg(
                        future_flags(frame.f_code), args, kwargs, 3, kw="flags"
                    )
                    args, kwargs = set_arg(1, args, kwargs, 4, kw="dont_inherit")
            elif flags & ast.PyCF_ONLY_AST:
                # If the code is being parsed to an AST, it can't directly be exec'd or eval'd
                # so we don't need to apply the rewriter. We also don't want to conservatively
                # rewrite, because other tools might use the AST and raise errors when
                # encountering Contrast rewritten nodes. See SUP-6215 for an example.
                return orig_func(*args, **kwargs)

            with scope.contrast_scope():
                code_to_compile, _ = rewrite(mode, code)
                args, kwargs = set_arg(code_to_compile, args, kwargs, 0, kw="source")
                result = orig_func(*args, **kwargs)
        except Exception:
            result = None
            raise
        finally:
            apply_rule(rule_applicator, orig_func, result, orig_args, orig_kwargs)

        return result

    return compile_patch


def patch_exec_and_eval(builtins_module):
    build_and_apply_patch(
        builtins_module,
        "eval",
        build_exec_eval_patch,
        (apply_unsafe_code_exec_rule, "eval"),
    )

    build_and_apply_patch(
        builtins_module,
        "exec",
        build_exec_eval_patch,
        (apply_unsafe_code_exec_rule, "exec"),
    )

    build_and_apply_patch(
        builtins_module, "compile", build_compile_patch, (apply_unsafe_code_exec_rule,)
    )


def register_patches():
    register_module_patcher(patch_exec_and_eval, builtins.__name__)


def reverse_patches():
    unregister_module_patcher(builtins.__name__)
    patch_manager.reverse_patches_by_owner(builtins)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
from functools import lru_cache
import os
import socket
import ssl
import sys

import logging as stdlib_logging
import logging.handlers as stdlib_logging_handlers
from typing import Any, NamedTuple, TextIO

import contrast
from contrast.assess_extensions import cs_str
from contrast.configuration.agent_config import AgentConfig
from contrast.configuration.config_option import ConfigOption
from contrast.utils.decorators import fail_loudly
from contrast.utils.loggers.structlog import RotationConfig, init_structlog
from contrast.utils.namespace import Namespace

from contrast.utils.string_utils import ensure_string
from contrast_vendor import structlog

from . import (
    DEFAULT_LOG_LEVEL,
    LOGGER_NAME,
    DEFAULT_AGENT_LOGGER_PATH,
)

STDOUT = "STDOUT"
STDERR = "STDERR"

CONFIG_KEY_STDOUT = "agent.logger.stdout"
CONFIG_KEY_STDERR = "agent.logger.stderr"
CONFIG_KEY_LOGGER_PATH = "agent.logger.path"
CONFIG_KEY_LOGGER_LEVEL = "agent.logger.level"


class module(Namespace):
    initialized: bool = False
    cef_security_logger: stdlib_logging.Logger | None = None
    syslog_logger: SysLogger | None = None


def setup_basic_agent_logger(level="INFO"):
    """
    Setup a logger without any user-supplied configuration, with defaults:
        1. log to stdout
        2. log in INFO level

    The logger created here is expected to be overridden with config values
    provided later on in the middleware creation cycle.
    """
    if not module.initialized:
        init_structlog(level, DEFAULT_AGENT_LOGGER_PATH)
        module.initialized = True

    return structlog.getLogger(LOGGER_NAME)


def configure_agent_logger(config: AgentConfig) -> None:
    """
    Configures the core agent logger with the user configuration.

    This function can be called multiple times to reconfigure the
    agent logger.
    """
    path_config_option, path = _logger_path(config)
    level_config_option = config.get_option(CONFIG_KEY_LOGGER_LEVEL)
    level = config[CONFIG_KEY_LOGGER_LEVEL].upper()

    cache_logger = _cache_logger(path_config_option, level_config_option)
    rotation_config = (
        None
        if (backups := config["agent.logger.backups"]) == 0
        else RotationConfig(
            backup_count=backups,
            # roll_size is in MB
            max_bytes=config["agent.logger.roll_size"] * 1024 * 1024,
        )
    )
    init_structlog(
        level,
        path,
        cache_logger=cache_logger,
        rotation_config=rotation_config,
    )

    cs_str.initialize_logger(structlog.getLogger(LOGGER_NAME))
    # Avoid circular import
    from contrast.agent.agent_lib import update_log_options

    update_log_options(level)


@lru_cache(maxsize=1)
@fail_loudly("Failed to get the fully qualified domain name.", return_value="-")
def _get_fqdn() -> str:
    return socket.getfqdn()


class CEFFormatter(stdlib_logging.Formatter):
    ATTACK_MSG_FMT = (
        f"%(asctime)s {socket.gethostname()} CEF:0|Contrast Security|Contrast Agent Python|{contrast.__version__}|"
        "SECURITY|%(message)s|%(level)s|pri=%(rule_id)s src=%(source_ip)s spt=%(source_port)s "
        "request=%(request_url)s requestMethod=%(request_method)s app=%(application)s "
        f"contrastAgentServer=%(server)s dvchost={_get_fqdn()} outcome=%(outcome)s"
    )

    def __init__(self):
        super().__init__(CEFFormatter.ATTACK_MSG_FMT, "%b %d %H:%M:%S")


def setup_security_logger(config: Mapping):
    if not config.get("protect.enable"):
        return

    path = config.get("agent.security_logger.path")
    level = config.get("agent.security_logger.level").upper()

    logger = stdlib_logging.getLogger("contrast-security-logger")

    _set_level(logger, level)

    handler = _get_handler(path)
    handler.setFormatter(CEFFormatter())
    logger.addHandler(handler)
    module.cef_security_logger = logger


def _unescaped_protect_rule_msg(rule_name, outcome, evaluation, masker):
    masked_input = masker.mask_attack_input(evaluation.input) if evaluation else None
    input_type = (
        evaluation.input.type.cef_string(masked_input.name) if masked_input else "-"
    )
    input_value = (
        masker.mask_attack_vector(ensure_string(masked_input.value, errors="replace"))
        if masked_input
        else "-"
    )
    if outcome == "exploited":
        if not evaluation:
            return f"An effective attack was detected against {rule_name}"
        return f"The {input_type} had a value that successfully exploited {rule_name} - {input_value}"
    if outcome in ("blocked", "ineffective"):
        if not evaluation:
            return f"An unsuccessful attack was detected against {rule_name}"
        return f"The {input_type} had a value that matched a signature for, but did not successfully exploit {rule_name} - {input_value}"
    if outcome == "suspicious":
        if not evaluation:
            return f"Suspicious activity indicates a potential attack using {rule_name} - {input_value}"
        return f"The {input_type} included a potential attack value that was detected as suspicious using {rule_name} - {input_value}"
    raise ValueError(f"Unknown outcome: {outcome}")


@fail_loudly("Failed to log attack event to security loggers")
def security_log_attack(attack, evaluation):
    """
    Logs a security event to the CEF security logger and syslog logger.

    Virtual patches, IP denylist and bot blocker activities are not yet supported.
    """
    if not module.cef_security_logger and not module.syslog_logger:
        # No security loggers configured.
        return

    from contrast.agent import agent_state

    app_name = agent_state.get_app_name()
    server_name = agent_state.get_server_name()
    masker = agent_state.module.request_data_masker

    ip = port = url = method_name = "-"

    if (context := contrast.REQUEST_CONTEXT.get()) is not None:
        ip = context.request.client_addr or "-"
        port = context.request.host_port
        method_name = context.request.method
        url = context.request.path
        masker = context.request_data_masker

    rule_name = attack.rule_id
    outcome = attack.report_result()
    msg = _escape_prefix(
        _unescaped_protect_rule_msg(rule_name, outcome, evaluation, masker)
    )

    log_context = _escape_metadata(
        {
            "level": "WARN",
            "rule_id": rule_name,
            "source_ip": ip,
            "source_port": port,
            "request_url": url,
            "request_method": method_name,
            "application": app_name or "-",
            "server": server_name or "-",
            "outcome": outcome.upper(),
        }
    )

    if module.cef_security_logger:
        module.cef_security_logger.warning(msg, extra=log_context)
    _syslog_msg(msg, log_context, outcome, perimeter=attack.perimeter_blocked)


def _escape_prefix(msg: str):
    # Order matters here. Escaping `|` before `\` would result in double escaping.
    return (
        msg.replace("\\", r"\\")
        .replace("|", r"\|")
        .replace("\n", r"\n")
        .replace("\r", r"\r")
    )


def _escape_metadata(metadata: dict[str, str]):
    return {
        # Order matters here. `\` must be escaped first since it's used to escape other characters.
        k: v.replace("\\", r"\\")
        .replace("=", r"\=")
        .replace("\n", r"\n")
        .replace("\r", r"\r")
        for k, v in metadata.items()
    }


def _syslog_msg(msg, log_context, outcome, perimeter):
    syslogger = module.syslog_logger
    if not syslogger:
        # syslog logger isn't configured.
        return
    outcome = outcome.lower()
    if perimeter and outcome == "blocked":
        outcome = "blocked_perimeter"
    syslogger.logger.warning(
        msg, extra={**log_context, "_severity": syslogger.outcome_to_severity[outcome]}
    )
    structlog.getLogger(LOGGER_NAME).debug("Sent security event to syslog")


class SysLogger(NamedTuple):
    logger: stdlib_logging.Logger
    outcome_to_severity: dict[str, str]


class SecureSysLogHandler(stdlib_logging_handlers.SysLogHandler):
    """
    A SysLogHandler that logs security events to syslog.

    This handler extends SysLogHandler to use SSL/TLS if the secure option is
    set to True. It also adds security event outcomes as log levels in the
    standard logging module.
    """

    _valid_severities = {
        "ALERT",
        "CRITICAL",
        "ERROR",
        "WARNING",
        "NOTICE",
        "INFO",
        "DEBUG",
    }

    @classmethod
    def severity(cls, s: str):
        if s.upper() not in cls._valid_severities:
            raise ValueError(f"Invalid syslog severity: {s}")
        return s.lower()

    _valid_protocols = {"UDP", "TCP", "TCP_TLS"}

    @classmethod
    def protocol(cls, p: str):
        if p.upper() not in cls._valid_protocols:
            raise ValueError(f"Invalid syslog protocol: {p}")
        return p.upper()

    def __init__(self, *args, secure=False, **kwargs):
        # Store the config so we can reference it later and potentially
        # avoid reconfiguring the logger if the config hasn't changed.
        self.config = {
            "args": args,
            "kwargs": {
                **kwargs,
                "secure": secure,
            },
        }

        super().__init__(*args, **kwargs)
        self.append_nul = False
        if secure:
            ctx = ssl.create_default_context()
            self.socket = ctx.wrap_socket(self.socket, server_hostname=self.address[0])

    def format(self, record) -> str:
        # HACK: The SysLogHandler extension capabilities are based on
        # adding custom logging levels, which is very difficult to do
        # safely as a library. Levels need integers and there's always
        # a chance the integer we choose will conflict with another
        # library. We'd prefer to avoid introducing log levels.
        #
        # Instead, we use this format call as a hook at the start of
        # the emit method to overwrite the levelname since levelname
        # is the only way to set the message priority.
        record.levelname = record.__dict__["_severity"]
        return super().format(record)

    def mapPriority(self, levelName: str) -> str:
        # bypass the priorityMap conversion.
        return levelName

    def close(self) -> None:
        """
        Closes the socket.

        Adds safety checks for before Python 3.10 to account for unset
        socket crash: https://github.com/python/cpython/issues/82961
        """
        if sys.version_info >= (3, 10):
            return super().close()

        self.acquire()
        try:
            if sock := getattr(self, "socket", None):
                self.socket = None
                sock.close()
            stdlib_logging.Handler.close(self)
        finally:
            self.release()


DEFAULT_SYSLOG_PORTS = {
    "UDP": 514,
    "TCP": 601,
    "TCP_TLS": 6514,
}


def configure_syslog_logger(config: Mapping) -> bool:
    """
    Configures the syslog logger with AgentConfig. This function can be called
    multiple times to reconfigure the logger with new configurations. The
    return value indicates whether the logger was reconfigured.

    The syslog logger uses the standard logging module to send messages to an
    adapted SysLogHandler.
    """
    if not config.get("protect.enable") or not config.get(
        "agent.security_logger.syslog.enable"
    ):
        return close_syslog_logger()

    outcome_to_severity: dict[str, str] = {
        "blocked": SecureSysLogHandler.severity(
            config.get("agent.security_logger.syslog.severity_blocked")
        ),
        "exploited": SecureSysLogHandler.severity(
            config.get("agent.security_logger.syslog.severity_exploited")
        ),
        "ineffective": SecureSysLogHandler.severity(
            config.get("agent.security_logger.syslog.severity_probed")
        ),
        "blocked_perimeter": SecureSysLogHandler.severity(
            config.get("agent.security_logger.syslog.severity_blocked_perimeter")
        ),
        "suspicious": SecureSysLogHandler.severity(
            config.get("agent.security_logger.syslog.severity_suspicious")
        ),
    }
    new_handler_config = _syslog_handler_config(config)

    logger, changed = _configure_syslog_std_logger(new_handler_config)
    module.syslog_logger = SysLogger(logger, outcome_to_severity)
    return changed


def _configure_syslog_std_logger(config) -> tuple[stdlib_logging.Logger, bool]:
    agent_logger = structlog.getLogger(LOGGER_NAME)

    logger = (
        module.syslog_logger.logger
        if module.syslog_logger
        else stdlib_logging.getLogger("contrast-syslog-logger")
    )
    previous_handler = None
    if (
        logger.handlers
        and (handler := logger.handlers[0])
        and isinstance(handler, SecureSysLogHandler)
    ):
        if handler.config == {"args": tuple(), "kwargs": config}:
            # The configuration hasn't changed. Continue to use the existing
            # handler.
            agent_logger.debug("Syslog logger configuration unchanged", config=config)
            return logger, False

        # previous_handler will be closed and removed after the new handler
        # is added. This ordering prevents dropping messages.
        #
        # A message could be written to both the previous_handler and the
        # latest handler. This is arguably duplication, but we can counter
        # by calling it redundancy, and this implementation side-steps the
        # need for additional locking or other synchronization.
        #
        # We could be finer grained in the condition above so that we keep
        # the existing connection in the address hasn't changed, but syslog
        # configurations from TeamServer don't happen frequently (at most once
        # per server settings polling period), so we'll keep it simple.
        previous_handler = handler

    try:
        handler = SecureSysLogHandler(**config)
    except OSError as e:
        agent_logger.error(
            "Failed to connect to syslog server",
            error=str(e),
            config=config,
        )
        return logger, False

    handler.setFormatter(CEFFormatter())

    logger.addHandler(handler)
    if previous_handler:
        previous_handler.close()
        logger.removeHandler(previous_handler)

    agent_logger.info("Syslog logger configuration updated", config=config)
    return logger, True


def _syslog_handler_config(config: Mapping) -> dict[str, Any]:
    protocol = SecureSysLogHandler.protocol(
        config.get("agent.security_logger.syslog.protocol")
    )
    host = config.get("agent.security_logger.syslog.server_host")
    if not host:
        # ip is deprecated in favor of server_host
        # In the future, it would be nice if the deprecated config
        # could be encapsulated within the get method.
        host = config.get("agent.security_logger.syslog.ip")
    port = config.get("agent.security_logger.syslog.port") or DEFAULT_SYSLOG_PORTS.get(
        protocol
    )
    facility = config.get("agent.security_logger.syslog.facility")
    socket_type = socket.SOCK_DGRAM if protocol == "UDP" else socket.SOCK_STREAM
    return {
        "address": (host, port),
        "facility": facility,
        "socktype": socket_type,
        "secure": protocol == "TCP_TLS",
    }


def close_syslog_logger():
    """
    Closes the syslog logger if it is currently enabled.

    Returns False if the logger is already disabled.
    """
    syslogger = module.syslog_logger
    if not syslogger:
        return False
    module.syslog_logger = None
    handlers = syslogger.logger.handlers
    for handler in handlers:
        handler.close()
    handlers.clear()
    return True


def _logger_path(
    config: AgentConfig,
) -> tuple[ConfigOption | None, TextIO | str]:
    if (option := config.get_option(CONFIG_KEY_STDOUT)) and option.value():
        path = sys.stdout
    elif (option := config.get_option(CONFIG_KEY_STDERR)) and option.value():
        path = sys.stderr
    elif (option := config.get_option(CONFIG_KEY_LOGGER_PATH)) and option.value():
        path = option.value()
    else:
        path = DEFAULT_AGENT_LOGGER_PATH
    return option, path


def _cache_logger(
    path_config_option: ConfigOption | None,
    level_config_option: ConfigOption | None,
) -> bool:
    return bool(
        path_config_option
        and level_config_option
        and path_config_option.is_definitely_static()
        and level_config_option.is_definitely_static()
    )


def _get_handler(path):
    if path == STDOUT:
        handler = stdlib_logging.StreamHandler(sys.stdout)
    elif path == STDERR:
        handler = stdlib_logging.StreamHandler(sys.stderr)
    else:
        try:
            if dirname := os.path.dirname(path):
                os.makedirs(dirname, exist_ok=True)
            handler = stdlib_logging.FileHandler(path)
        except Exception as e:
            sys.stderr.write(f"{str(e)}\n")
            # path could be '' or None
            handler = stdlib_logging.StreamHandler()

    return handler


def _set_level(logger, level: str) -> None:
    if level.upper() == "TRACE":
        level = "DEBUG"
        sys.stderr.write(
            "Contrast Python Agent: TRACE logging is equivalent to DEBUG\n"
        )
    try:
        logger.setLevel(level)
    except ValueError:
        # this fails validation if the level is an invalid value
        logger.setLevel(DEFAULT_LOG_LEVEL)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import sys

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    unregister_module_patcher,
    wrap_and_watermark,
    register_module_patcher,
)
from contrast.agent.middlewares.route_coverage import common

from contrast_vendor import structlog as logging

DJANGO_WSGI_NAME = "django.core.wsgi"
DJANGO_ASGI_NAME = "django.core.asgi"


@fail_quietly("failed to apply django config scanning rules")
def do_config_scanning(logger):
    # Lazy import for django
    from django.conf import settings as app_settings
    from contrast.agent.assess.rules.config import (
        DjangoHttpOnlyRule,
        DjangoSecureFlagRule,
        DjangoSessionAgeRule,
    )

    logger.debug("applying django config scanning rules")
    app_config_module_name = os.environ.get("DJANGO_SETTINGS_MODULE")
    if not app_config_module_name:
        logger.warning("unable to find django settings for config scanning")
        return

    app_config_module = sys.modules.get(app_config_module_name)
    if not app_config_module:
        logger.warning("django settings module not loaded; can't scan config")
        return

    for rule in [DjangoSessionAgeRule, DjangoSecureFlagRule, DjangoHttpOnlyRule]:
        rule().apply(app_settings, app_config_module)


@fail_quietly("failed to run django first request analysis")
@scope.contrast_scope()
def django_first_request(sender, **kwargs):
    from django.core import signals
    from contrast.agent.middlewares.route_coverage.django_routes import (
        create_django_routes,
    )

    del sender, kwargs

    logger = logging.getLogger("contrast")
    logger.debug("called django first request signal")

    do_config_scanning(logger)
    common.handle_route_discovery("django", create_django_routes, ())

    # the signal is no longer needed after the first request
    logger.debug("disconnecting django first request signal (first request started)")
    signals.request_started.disconnect(
        django_first_request, dispatch_uid="contrast_first_request"
    )


def _extract_path(signal_kwargs):
    """Extract path based on wsgi environ vs asgi scope"""
    # Handles the wsgi case
    environ = signal_kwargs.get("environ")
    if environ is not None:
        return environ.get("PATH_INFO", "")

    # Handles the asgi case
    scope = signal_kwargs.get("scope")
    if scope is not None:
        return scope.get("path", "")

    return ""


@fail_quietly("failed to run django request analysis")
@scope.contrast_scope()
def django_every_request(sender, **kwargs):
    """
    Django signal handler that performs route observation

    We expect the `environ` kwarg (for wsgi) or the `scope` kwarg (for asgi) to
    be present and not None. If these conditions do not hold, they will be
    handled by `fail_quietly`.
    """
    # Lazy import for django
    from contrast.agent.middlewares.route_coverage.django_routes import (
        build_django_signature,
        get_matched_resolver,
    )

    del sender

    logger = logging.getLogger("contrast")
    logger.debug("performing django route observation")

    context = contrast.REQUEST_CONTEXT.get()
    if context is None:
        logger.debug("not in request context - skipping route observation")
        return

    request_path = _extract_path(kwargs)
    resolved = get_matched_resolver(request_path)
    if resolved is None:
        logger.debug("did not find django view function for route observation")
        return

    context.signature = build_django_signature(resolved)
    if resolved.route is not None:
        context.path_template = resolved.route
    logger.debug(
        "Observed route",
        signature=context.signature,
        path_template=context.path_template,
    )


@fail_quietly("Failed to retrieve django application name from settings")
def get_app_name() -> str:
    # Lazy import for django
    from django.conf import settings

    application = getattr(
        settings,
        "WSGI_APPLICATION",
        "",
    ) or getattr(
        settings,
        "ASGI_APPLICATION",
        "",
    )

    return application.split(".")[0]


@fail_loudly("Failed to initialize Django-specific instrumentation")
def initialize_django():
    from contrast.agent.agent_state import set_application_name

    # Lazy import for django
    from django.core import signals

    logger = logging.getLogger("contrast")
    logger.info("Detected application: django")

    # TODO: PYT-2852 Revisit application name detection
    set_application_name(get_app_name())

    logger.debug("Connecting django signals")
    signals.request_started.connect(
        django_every_request, dispatch_uid="contrast_every_request"
    )
    # This signal runs on the first request only
    # Runs at the end of the first request to make sure we gather all registered routes
    signals.request_started.connect(
        django_first_request, dispatch_uid="contrast_first_request"
    )


def disconnect_django():
    from django.core import signals

    signals.request_started.disconnect(
        django_every_request, dispatch_uid="contrast_every_request"
    )
    signals.request_started.disconnect(
        django_first_request, dispatch_uid="contrast_first_request"
    )


def build_get_wsgi_app_patch(orig_func, _):
    def get_wsgi_application(wrapped, _, args, kwargs):
        # Avoids circular import
        from contrast.agent.agent_state import automatic_middleware
        from contrast.wsgi import ContrastMiddleware

        # This call needs to occur before middleware initialization
        initialize_django()

        with automatic_middleware():
            return ContrastMiddleware(wrapped(*args, **kwargs), framework_name="django")

    return wrap_and_watermark(orig_func, get_wsgi_application)


def build_get_asgi_app_patch(orig_func, _):
    def get_asgi_application(wrapped, _, args, kwargs):
        # Avoids circular import
        from contrast.agent.agent_state import automatic_middleware
        from contrast.asgi import ContrastMiddleware

        # This call needs to occur before middleware initialization
        initialize_django()

        with automatic_middleware():
            return ContrastMiddleware(wrapped(*args, **kwargs), framework_name="django")

    return wrap_and_watermark(orig_func, get_asgi_application)


def patch_django_wsgi(module):
    # XXX: It may be better to hook django.core.handlers.base.BaseHandler.__init__
    # It's possible that this will be more general solution for other servers
    build_and_apply_patch(module, "get_wsgi_application", build_get_wsgi_app_patch)


def patch_django_asgi(module):
    build_and_apply_patch(module, "get_asgi_application", build_get_asgi_app_patch)


def register_patches():
    register_module_patcher(patch_django_wsgi, DJANGO_WSGI_NAME)
    register_module_patcher(patch_django_asgi, DJANGO_ASGI_NAME)


def reverse_patches():  # pragma: no cover
    unregister_module_patcher(DJANGO_WSGI_NAME)
    unregister_module_patcher(DJANGO_ASGI_NAME)
    patch_manager.reverse_module_patches_by_name(DJANGO_WSGI_NAME)
    patch_manager.reverse_module_patches_by_name(DJANGO_ASGI_NAME)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from functools import cached_property

import contrast

from aiohttp.web import StreamResponse
from aiohttp.web_urldispatcher import DynamicResource

from contrast.agent.request_context import RequestContext
from contrast.aiohttp import sources
from contrast.agent import scope, request_state, agent_state
from contrast.agent.middlewares.route_coverage.aiohttp_routes import (
    create_aiohttp_routes,
)
from contrast.agent.middlewares.route_coverage import common
from contrast.agent.policy.trigger_node import TriggerNode
from contrast_vendor import structlog as logging
from contrast.agent.middlewares.base_middleware import (
    BaseMiddleware,
)
from contrast.agent.middlewares.response_wrappers.aiohttp_response_wrapper import (
    AioHttpResponseWrapper,
)

from contrast.utils.decorators import fail_quietly

logger = logging.getLogger("contrast")


class AioHttpMiddleware(BaseMiddleware):
    __middleware_version__ = 1  # Aiohttp new-style middleware

    # Since there is no way to get the `app` instance, on startup of AioHttp,
    # until the first request, hence we will not have `app` finder logic.
    @scope.contrast_scope()
    def __init__(self, app_name: str | None = None) -> None:
        self.app = None
        self.app_name = app_name or "aiohttp"
        agent_state.set_detected_framework("aiohttp")
        super().__init__()

    async def __call__(self, request, handler) -> StreamResponse:
        if request_state.get_request_id() is not None:
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request_id - passing through")
            return await handler(request)

        self.app = request.app

        with self.request_context_stack(
            request.method, request.path
        ) as request_context_stack:
            with scope.contrast_scope():
                environ = await sources.aiohttp_request_to_environ(request)

            context = self.should_analyze_request(environ)
            if context:
                self.enter_agent_call_contexts(request_context_stack, context)
                return await self.call_with_agent(context, request, handler)

            return await self.call_without_agent_async(request, handler)

    async def call_with_agent(
        self, context: RequestContext, request, handler
    ) -> StreamResponse:
        with scope.contrast_scope():
            sources.track_aiohttp_request_sources(context, request)

            try:
                self.prefilter(context)

                with scope.pop_contrast_scope():
                    try:
                        response = await handler(request)
                        wrapped_response = AioHttpResponseWrapper(response)
                        context.extract_response(wrapped_response)
                    except Exception as exc:
                        context.response_exception = exc
                        raise

                self.postfilter(context)
                self.check_for_blocked(context)

                return response

            finally:
                if context.assess_enabled:
                    self.do_aiohttp_first_request_analysis()
                    self.do_aiohttp_route_observation(context, request)
                self.handle_ensure(context, request)
                if context.assess_enabled:
                    contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, request, handler) -> StreamResponse:
        with scope.contrast_scope():
            return await handler(request)

    @fail_quietly()
    def do_aiohttp_first_request_analysis(self) -> None:
        if not agent_state.is_first_request():
            return

        common.handle_route_discovery("aiohttp", create_aiohttp_routes, (self.app,))

    @fail_quietly()
    def do_aiohttp_route_observation(self, context, request) -> None:
        view_func = self.get_aiohttp_view_func(request)
        if view_func is None:
            logger.debug("unable to get view function for aiohttp route observation")
            return

        context.signature = common.build_signature(view_func.__name__, view_func)
        logger.debug("Observed aiohttp route", signature=context.signature)

    @fail_quietly("Unable to get view func")
    def get_aiohttp_view_func(self, request):
        """
        This intentionally does not override get_view_func. We're generally making route
        coverage more framework-specific; this minimizes shared machinery between
        aiohttp middleware and its base classes.
        """
        if not request.path:
            return None

        view_func = None

        # This approach has at worst O(_resources) performance
        # but it's a first attempt at implementing a sync
        # version of aiohttp.web_urldispatches.UrlDispatcher.resolve
        for app_route in self.app.router._resources:
            _app_route = (
                app_route._formatter
                if isinstance(app_route, DynamicResource)
                else app_route._path
            )
            if _app_route == request.path:
                routes = app_route._routes
                for route in routes:
                    if route.method == request.method:
                        return route.handler

        return view_func

    @cached_property
    def trigger_node(self):
        """
        Used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "aiohttp"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import collections
import itertools
import sys
import traceback
from functools import lru_cache
from types import FrameType

from contrast import AGENT_CURR_WORKING_DIR
from contrast.utils.decorators import fail_quietly
from contrast_fireball import AssessStackFrame
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


APPLIES_MARKER = "cs__"
PATCH_MARKER = "__cs"
PY_FILE_EXTENSION = ".py"
UTILS_MODULES = "contrast/utils"
NONETYPE = "NoneType"

CONTRAST_EXTENSIONS = (
    "contrast_vendor",
    "contrast_rewriter",
    "contrast/assess_extensions",
    "contrast/patches",
)
DJANGO_EXCEPTION_PATH = "core/handlers/exception.py"
DJANGO_DEPRECATION_PATH = "utils/deprecation.py"
SITE_PACKAGES = "site-packages"

# PERF: limit the number of stack frames to not use as much memory.
STACK_LIMIT = 20


@fail_quietly("Failed to build stacktrace for event", return_value=[])
def build_stack(limit=STACK_LIMIT) -> StackSummary:
    return extract_stack(limit=limit)


@fail_quietly("Failed to clean stacktrace for event", return_value=[])
def clean_stack(frames: StackSummary):
    return [to_element(x) for x in reversed(frames)]


@fail_quietly("Failed to convert event")
def to_element(summary: traceback.FrameSummary):
    from contrast.reporting.reporting_client import StackFrame

    return StackFrame(
        line_number=summary.lineno,
        file_name=filename_formatter(summary.filename),
        declaring_class=summary.filename,
        method_name=summary.name,
    )


@fail_quietly("Failed to clean assess stacktrace", return_value=[])
def to_assess_stack(summary: StackSummary):
    return [
        AssessStackFrame(
            line_number=frame.lineno or -1,
            file=filename_formatter(frame.filename),
            method=frame.name,
        )
        for frame in reversed(summary)
    ]


def extract_stack(f: FrameType | None = None, limit=None):
    """Extract the raw traceback from the current stack frame.

    Instrumentation frames are filtered out of the results and are
    not included in the calculation of limits.

    The return value has the same format as for extract_tb().  The
    optional 'f' and 'limit' arguments have the same meaning as for
    print_stack().  Each item in the list is a quadruple (filename,
    line number, function name, text), and the entries are in order
    from oldest to newest stack frame.
    """
    if f is None:
        f = sys._getframe().f_back
    if f is None:
        return StackSummary()
    stack = StackSummary.extract(walk_stack(f), limit=limit, lookup_lines=False)
    stack.reverse()
    return stack


def walk_stack(f: FrameType | None):
    """Walk a stack yielding the frame and line number for each frame.

    This will follow f.f_back from the given frame.

    Unlike traceback.walk_stack, if no frame is given, no frames are
    yielded.

    It will also filter out frames belonging to Contrast instrumentation.
    """
    while f is not None:
        if acceptable_frame(f):
            yield f, f.f_lineno
        f = f.f_back


def acceptable_frame(frame: FrameType):
    """
    Return true if frame does NOT belong to Contrast instrumentation.
    """

    filename = frame.f_code.co_filename
    name = frame.f_code.co_name
    return (
        "/contrast/" not in filename
        and UTILS_MODULES not in filename
        and not any(extension in filename for extension in CONTRAST_EXTENSIONS)
        and not name.startswith(APPLIES_MARKER)
        and not name.startswith(PATCH_MARKER)
        and not filename.endswith(DJANGO_EXCEPTION_PATH)
        and not filename.endswith(DJANGO_DEPRECATION_PATH)
    )


class StackSummary(list[traceback.FrameSummary]):
    """A stack of frames."""

    @classmethod
    def extract(cls, frame_gen, *, limit=None, lookup_lines=True, capture_locals=False):
        """Create a StackSummary from a traceback or stack object.

        Like traceback.StackSummary.extract, but without linecache to avoid
        os.stat calls and improve performance.

        :param frame_gen: A generator that yields (frame, lineno) tuples to
            include in the stack.
        :param limit: None to include all frames or the number of frames to
            include.
        :param lookup_lines: If True, lookup lines for each frame immediately,
            otherwise lookup is deferred until the frame is rendered.
        :param capture_locals: If True, the local variables from each frame will
            be captured as object representations into the FrameSummary.
        """
        if limit is None:
            limit = getattr(sys, "tracebacklimit", None)
            if limit is not None and limit < 0:
                limit = 0
        if limit is not None:
            if limit >= 0:
                frame_gen = itertools.islice(frame_gen, limit)
            else:
                frame_gen = collections.deque(frame_gen, maxlen=-limit)

        result = cls()
        fnames = set()
        for f, lineno in frame_gen:
            co = f.f_code
            filename = co.co_filename
            name = co.co_name

            fnames.add(filename)
            # Must defer line lookups until we have called checkcache.
            f_locals = f.f_locals if capture_locals else None
            result.append(
                traceback.FrameSummary(
                    filename, lineno, name, lookup_line=False, locals=f_locals
                )
            )
        # If immediate lookup was desired, trigger lookups now.
        if lookup_lines:
            for f in result:
                _ = f.line
        return result


def build_and_clean_stack(depth=STACK_LIMIT):
    """
    Perform both build and clean steps.
    """
    return clean_stack(build_stack(limit=depth))


SORTED_FILENAME_SEARCH_PATH = sorted(
    set(sys.path) | {AGENT_CURR_WORKING_DIR}, key=len, reverse=True
)


@fail_quietly("Unable to create file_name")
@lru_cache(maxsize=512)
def filename_formatter(file_name: str):
    # PERF: This method is called hundreds of times, so be mindful
    # of what additional computations are added.

    if file_name.startswith("<frozen"):
        return file_name

    for sys_path in SORTED_FILENAME_SEARCH_PATH:
        if file_name.startswith(sys_path):
            file_name = file_name.removeprefix(sys_path)
            break

    return file_name.replace("/", ".").lstrip(".")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from functools import cached_property
from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.policy.constants import ALL_ARGS, ALL_KWARGS, OBJECT
from contrast.agent.policy.policy_node import PolicyNode
from contrast_fireball import AssessEventType


class PropagationNode(PolicyNode):
    TAGGER = "Tagger"
    PROPAGATOR = "Propagator"

    def __init__(
        self,
        module,
        class_name,
        instance_method,
        method_name,
        source,
        target,
        action,
        tags=None,
        untags=None,
        policy_patch=True,
    ):
        """
        :param class_name: class of the hook
        :param instance_method: if the method is an instance method vs static/bound/unbound
        :param method_name: method name to hook
        :param source: from where the tainted data flows, cannot be None
        :param target: to where the tainted data flows, cannot be None
        :param action: how the tainted data flows from source to target, should not be None
        :param tags: array of tags to apply to the target, can be None if no tags are added
        :param untags: array of tags to remove from the target, can be None if not tags are removed
        """
        super().__init__(
            module,
            class_name,
            instance_method,
            method_name,
            source,
            target,
            tags,
            policy_patch=policy_patch,
        )

        self.action = action
        self.untags = set(untags) if untags else set()

        self.validate()

    @property
    def node_class(self):
        return self.TAGGER if self.is_tagger else self.PROPAGATOR

    @property
    def node_type(self):
        return AssessEventType.TAG if self.is_tagger else AssessEventType.PROPAGATION

    @cached_property
    def is_tagger(self):
        return bool(self.tags) or bool(self.untags)

    def validate(self):
        super().validate()

        if not (self.targets and len(self.targets) != 0):
            raise ContrastAssessException(
                f"Propagator {self.method_name} did not have a proper target. Unable to create."
            )

        if not (self.sources and len(self.sources) != 0):
            raise ContrastAssessException(
                f"Propagator {self.method_name} did not have a proper source. Unable to create."
            )

        if not self.action:
            raise ContrastAssessException(
                f"Propagator {self.method_name} did not have a proper action. Unable to create."
            )

        self.validate_untags()

    def validate_untags(self):
        if not self.untags:
            return

        for item in self.untags:
            if item not in constants.VALID_TAGS:
                raise ContrastAssessException(
                    f"{self.node_type} {self.id} did not have a valid untag. {item} is not a known value."
                )

            if self.tags and item in self.tags:
                raise ContrastAssessException(
                    f"{self.node_type} {self.id} had the same tag and untag, {item}."
                )

    def get_matching_sources(self, preshift):
        sources = []
        args = (
            preshift.args[1:]
            # The string propagation hooks do not pass `self` as part of the args tuple
            # so we need to account for that here.
            if preshift.args and self.class_name != "str" and self.instance_method
            else preshift.args
        )

        for source in self.sources:
            if source == OBJECT:
                sources.append(preshift.obj)
            elif source == ALL_ARGS:
                sources.extend(args)
            elif source == ALL_KWARGS:
                sources.append(preshift.kwargs)
            elif isinstance(source, int) and len(args) > source:
                sources.append(args[source])
            elif preshift.kwargs and source in preshift.kwargs:
                sources.append(preshift.kwargs[source])

        return sources



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent import scope
from contrast.agent.patch_controller import (
    enable_assess_patches,
    is_preinstrument_flag_set,
)
from contrast.agent.policy.rewriter import apply_rewrite_policy
from contrast.configuration.agent_config import AgentConfig
from contrast.patches import (
    register_chaining_monkeypatches,
    register_automatic_middleware_monkeypatches,
)

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


@scope.contrast_scope()
def apply_early_instrumentation():
    """
    Applies a subset of instrumentation during site loading.

    Currently, only instrumentation that needs to be applied as early as possible
    is applied here. Eventually, it would be simpler to apply all instrumentation,
    but that will require more refactoring.
    """

    preinstrument = is_preinstrument_flag_set()

    config = AgentConfig()

    # Policy-based rewrites need to be applied prior to any policy patches.
    # Policy patches can be layered on top of rewritten functions. So that
    # means we need to make sure that the "original" function called by the
    # policy patch is the *rewritten* one.
    if config.assess_enabled or preinstrument:
        logger.info("Applying policy-based rewrites")
        apply_rewrite_policy()

    if config.enable_automatic_middleware or preinstrument:
        logger.info("Registering automatic instrumentation patches")
        register_automatic_middleware_monkeypatches()

    if config.assess_enabled or preinstrument:
        logger.info("Registering foundational Assess patches")
        # NOTE: policy is currently loaded/generated on import. It is applied explicitly
        # in policy/applicator.py later
        from contrast import policy  # noqa: F401

        enable_assess_patches()

    logger.info("Registering instrumentation chaining patches")
    register_chaining_monkeypatches()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

import os
import pkgutil
import sys
from collections.abc import Generator, Iterable
from zlib import crc32

from contrast.utils.libraries import distributions, normalize_dist_name, Distribution
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


def artifact_fingerprint() -> str:
    """
    Compute a fingerprint of the running code artifact.

    This fingerprint is smudgy. It prefers to be fast and simple over being
    completely accurate.

    It may return an empty string if it fails to compute a meaningful fingerprint.

    The returned fingerprint is based on the installed distribution packages,
    their versions, and the names and sized of the application source files.
    """
    app_root = sys.path[0] or "."
    app_ids = app_source_ids(app_root)
    if not app_ids:
        logger.debug("failed to compute fingerprint, no app sources found")
        return ""
    deps_ids = distribution_ids(distributions())
    artifact_id = f"app_sources={app_ids};deps={deps_ids}"
    logger.debug(
        "computed fingerprint",
        artifact_id=artifact_id,
        app_sources_count=len(app_ids),
        installed_distributions_count=len(deps_ids),
    )
    return str(crc32(artifact_id.encode()))


def distribution_ids(dists: Iterable[Distribution]) -> list[str]:
    return sorted(
        [
            f"{normalize_dist_name(dist_meta['Name'])}=={dist_meta['Version']}"
            for dist_meta in (dist.metadata for dist in dists)
        ]
    )


def app_source_ids(app_root: str) -> list[str]:
    return [
        module_id(info, origin)
        for info, origin in sorted(
            scan_app_modules([app_root], prefix="."), key=lambda x: x[0].name
        )
    ]


def module_id(info: pkgutil.ModuleInfo, origin: str):
    file_size = os.stat(origin).st_size
    return f"{info.name}:{file_size}"


def scan_app_modules(
    search_locations: list[str] | None, prefix=""
) -> Generator[tuple[pkgutil.ModuleInfo, str], None, None]:
    if not search_locations:
        return
    for mod_info in pkgutil.iter_modules(search_locations, prefix):
        spec = mod_info.module_finder.find_spec(mod_info.name)
        if not spec or not spec.origin:
            continue
        yield mod_info, spec.origin
        yield from scan_app_modules(
            spec.submodule_search_locations, prefix=f"{mod_info.name}."
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from enum import Enum, auto

import contrast
from contrast.agent.request import Request
from contrast.api.sample import Sample
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ProtectResponse(Enum):
    NO_ACTION = auto()
    BLOCKED = auto()
    MONITORED = auto()
    PROBED = auto()
    BLOCKED_AT_PERIMETER = auto()


# Certain rules in Protect can only be confirmed as suspicious, meaning they didn't get evaluated against user input
# or they didn't have input tracing applied. We report these rules differently.
SUSPICIOUS_RULES = [
    "malformed-header",
    "reflected-xss",
    "unsafe-file-upload",
    "zip-file-overwrite",
]


class Attack:
    """
    Class storing all data necessary to report a protect attack.
    """

    def __init__(self, rule_id: str):
        self.rule_id: str = rule_id
        self.samples: list[Sample] = []
        self.response: ProtectResponse | None = None
        self.start_time_ms = contrast.REQUEST_CONTEXT.get().request.timestamp_ms

    @property
    def blocked(self) -> bool:
        return self.response == ProtectResponse.BLOCKED

    @property
    def perimeter_blocked(self) -> bool:
        return self.response == ProtectResponse.BLOCKED_AT_PERIMETER

    def add_sample(self, sample: Sample) -> None:
        self.samples.append(sample)

    def _convert_samples(self, request: Request) -> list[dict]:
        reportable_request = request.reportable_format if request is not None else {}

        return [
            {
                "blocked": self.blocked,
                "input": self._convert_user_input(sample),
                "details": self._mask_details(sample),
                "request": reportable_request,
                "stack": [
                    {
                        "declaringClass": stack.declaring_class,
                        "methodName": stack.method_name,
                        "fileName": stack.file_name,
                        "lineNumber": stack.line_number,
                    }
                    for stack in sample.stack
                ],
                "timestamp": {
                    "start": sample.timestamp_ms,  # in ms
                    "elapsed": (
                        now_ms() - sample.timestamp_ms
                    ),  # in ms which is the format TS accepts
                },
            }
            for sample in self.samples
        ]

    def _mask_details(self, sample: Sample) -> dict:
        if not self.masker.mask_rules.mask_attack_vector:
            return sample.details

        return {
            k: self.masker.mask_attack_vector(v) if isinstance(v, str) else v
            for k, v in sample.details.items()
        }

    def _convert_user_input(self, sample: Sample) -> dict:
        masked_user_input = self.masker.mask_attack_input(sample.user_input)
        json_sample = {
            "documentType": masked_user_input.document_type.name,
            "filters": masked_user_input.matcher_ids,
            "time": sample.timestamp_ms,
            "type": masked_user_input.type.name,
            "value": masked_user_input.value,
        }
        if masked_user_input.path:
            json_sample["documentPath"] = masked_user_input.path
        if masked_user_input.name:
            json_sample["name"] = masked_user_input.name

        return json_sample

    def report_result(self) -> str:
        response_map = {
            ProtectResponse.BLOCKED: "blocked",
            ProtectResponse.BLOCKED_AT_PERIMETER: "blocked",
            ProtectResponse.PROBED: "ineffective",
            ProtectResponse.MONITORED: (
                "suspicious" if self.rule_id in SUSPICIOUS_RULES else "exploited"
            ),
        }
        return response_map[self.response]

    def to_json(self, request: Request, request_data_masker) -> dict:
        self.masker = request_data_masker
        request_data_masker.mask_sensitive_data(request, self)

        common_fields = {
            "startTime": 0,
            "total": 0,
        }
        json = {
            "startTime": self.start_time_ms,
            "blocked": common_fields,
            "exploited": common_fields,
            "ineffective": common_fields,
            "suspicious": common_fields,
        }

        relevant_mode = self.report_result()

        samples = self._convert_samples(request)

        json[relevant_mode] = {
            "total": 1,  # always 1 until batching happens
            "startTime": self.start_time_ms,
            "samples": samples,
        }

        return json



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class _NoConstructorArg:
    pass


class TypeCheckedProperty:
    """
    A property that performs a type-check when set. If `__get__` is called before
    `__set__`, a value is created using the type's default constructor. This sort of
    mimics protobuf behavior.

    Note that type-checking is only performed on the top-level type. There is no
    recursive checking for nested data structures.
    """

    def __init__(self, _type, *, constructor_arg: object = _NoConstructorArg):
        self._type = _type
        self._constructor_args = (
            [] if constructor_arg is _NoConstructorArg else [constructor_arg]
        )

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, objtype=None):
        if self.name not in obj.__dict__:
            self.__set__(obj, self._type(*self._constructor_args))
        return obj.__dict__[self.name]

    def __set__(self, obj, value):
        if not isinstance(value, self._type):
            raise TypeError(
                f"Bad type when setting TypeCheckedProperty "
                f'"{self.name}" on {type(obj)} - '
                f"expected: {self._type}, found: {type(value)}"
            )
        obj.__dict__[self.name] = value



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import inspect
import sys
from contextvars import ContextVar
from collections.abc import Iterable
from typing import TYPE_CHECKING

from contrast_fireball import DiscoveredRoute
from contrast.agent.assess.policy.source_policy import cs__apply_source
from contrast_vendor import wrapt

import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy import registry
from contrast.agent.policy import patch_manager
from contrast.agent.middlewares.route_coverage import common
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)

if TYPE_CHECKING:
    from starlette.routing import Router, Route
    from starlette.types import ASGIApp

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

DEFAULT_STARLETTE_ROUTE_METHODS = common.DEFAULT_ROUTE_METHODS + ("HEAD",)

STARLETTE_ROUTING = "starlette.routing"
STARLETTE_REQUESTS = "starlette.requests"


def build___call___patch(orig_func, patch_policy):
    """
    Patch for starlette.routing.Router.__call__

    This currently gives us route discovery and route observation. If in the future we
    need a reference to the actual FastAPI / Starlette application object, we'll need
    another patch (probably for one of those class's __call__ methods).

    The innermost application object in any Starlette-based application is a Router.
    The router is a valid ASGI application object. We cannot patch
    starlette.applications.Starlette.__call__ (or fastapi.FastAPI.__call__) because of
    the order in which these methods are invoked during request processing. Starlette's
    unique middleware installation procedure leads to the following call order:

    fastapi.FastAPI.__call__
      starlette.applications.Starlette.__call__
        each middleware's __call__
          starlette.routing.Router.__call__

    This differs from typical middleware installation, where the middleware __call__
    methods come first.

    We need to perform route discovery / observation while we're still somewhere inside
    of the agent middleware's __call__ method; otherwise we won't have this information
    in time for handle_ensure() at the end of agent request processing. The best option
    seems to be Router.__call__.
    """
    del patch_policy

    context = ContextVar("starlette_router_call_context", default=False)

    async def __call___patch(wrapped, instance, args, kwargs):
        if context.get():
            # Avoid reporting route coverage from inner Routers,
            # possible from mounting sub-applications.
            return await wrapped(*args, **kwargs)
        token = context.set(True)
        # PERF: avoid extra try/finally while manipulating context var.
        # We want to do this because this patch is hit on every request.
        # We can safely do this because route discovery and observation
        # functions are already protected with fail_quietly. There's also
        # a unit test to ensure this.
        do_starlette_route_discovery(instance)
        try:
            result = await wrapped(*args, **kwargs)
        finally:
            do_starlette_route_observation(instance, *args, **kwargs)
            context.reset(token)
        return result

    return wrap_and_watermark(orig_func, __call___patch)


@fail_quietly("Failed to run starlette first-request analysis")
@scope.contrast_scope()
def do_starlette_route_discovery(starlette_router_instance: Router):
    from contrast.agent import agent_state

    if not agent_state.is_first_request():
        return

    common.handle_route_discovery(
        "starlette", create_starlette_routes, (starlette_router_instance,)
    )


def create_starlette_routes(asgi_app: ASGIApp, prefix="") -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a Starlette router.
    """
    from starlette.routing import Host, Mount, Router, Route, WebSocketRoute
    from starlette.staticfiles import StaticFiles

    if isinstance(asgi_app, Router):
        routes = set()
        for app_route in asgi_app.routes:
            routes.update(create_starlette_routes(app_route, prefix=prefix))
        return routes

    if isinstance(asgi_app, Route):
        app_route = asgi_app
        view_func = app_route.endpoint

        signature = common.build_signature(app_route.name, view_func)
        path_template = app_route.path
        methods = starlette_route_methods(app_route)
        if methods is None:
            # If methods is None, it means all methods are allowed.
            # The backend considers a missing method to mean "any method",
            methods = [None]

        return {
            DiscoveredRoute(
                verb=method_type,
                url=prefix + path_template,
                signature=signature,
                framework="Starlette",
            )
            for method_type in methods
        }

    if isinstance(asgi_app, StaticFiles):
        # StaticFiles aren't registered as routes on the Router, but we still
        # want to report them. They return 405 for non-GET/HEAD requests.
        return {
            DiscoveredRoute(
                verb="GET",
                url=prefix + "/{path:path}",
                signature=f"StaticFiles(directory={asgi_app.directory})",
                framework="Starlette",
            ),
            DiscoveredRoute(
                verb="HEAD",
                url=prefix + "/{path:path}",
                signature=f"StaticFiles(directory={asgi_app.directory})",
                framework="Starlette",
            ),
        }

    if isinstance(asgi_app, Mount):
        return create_starlette_routes(asgi_app.app, prefix=prefix + asgi_app.path)
    if isinstance(asgi_app, Host):
        return create_starlette_routes(asgi_app.app, prefix=prefix)

    if isinstance(asgi_app, WebSocketRoute):
        # We don't support instrumenting WebSockets, so we don't report them.
        return set()

    # Fallback for any other ASGI app.

    view_name = asgi_app.__module__
    if not inspect.isfunction(asgi_app):
        view_name += f".{asgi_app.__class__.__qualname__}"
    else:
        view_name += f".{asgi_app.__qualname__}"

    signature = common.build_signature(view_name, asgi_app)
    return {
        DiscoveredRoute(
            verb=None,  # We don't know the allowed methods here, so fail open.
            url=prefix + "/{path:path}",
            signature=signature,
            framework="Starlette",
        )
    }


def starlette_route_methods(route: Route) -> Iterable[str] | None:
    """
    Returns the allowed HTTP methods for a given Starlette Route.
    If the route allows all methods, returns None.
    """
    from starlette.endpoints import HTTPEndpoint

    methods = route.methods
    if inspect.isclass(route.endpoint) and issubclass(route.endpoint, HTTPEndpoint):
        # Starlette is very permissive about HTTP method routing. If the HTTP method
        # (a.k.a. verb) matches a method on the endpoint class, that method will be
        # called. If not, the endpoint's method_not_allowed() method will be called.
        # This means that internal methods like _some_helper could be called if a
        # request was sent with HTTP method "_SOME_HELPER". In most cases, reporting
        # this would be noise to the user because these methods might have different
        # calling conventions (e.g. not accepting a request parameter or requiring
        # other parameters), and sending a request with a non-standard HTTP method
        # will result in a 500 Internal Server Error from unhandled TypeErrors.
        # To avoid this noise, we only report methods that are explicitly defined on
        # the endpoint class with the signature suggested in the Starlette docs:
        #     https://starlette.dev/endpoints/#httpendpoint
        allowed_methods = [
            name.upper()
            for (name, member) in inspect.getmembers(
                route.endpoint,
                predicate=(inspect.isfunction),
            )
            if name != "method_not_allowed"
            and (
                list(inspect.signature(member).parameters.keys()) == ["self", "request"]
            )
        ]
        if methods is None:
            methods = allowed_methods
        else:
            methods.intersection_update(allowed_methods)
    return methods


@fail_quietly("unable to perform starlette route observation")
@scope.contrast_scope()
def do_starlette_route_observation(
    starlette_router_instance, asgi_scope, *args, **kwargs
):
    from starlette.endpoints import HTTPEndpoint
    from starlette.staticfiles import StaticFiles

    context = contrast.REQUEST_CONTEXT.get()
    if context is None:
        return

    logger.debug("Performing starlette route observation")

    if not asgi_scope or not isinstance(asgi_scope, dict):
        logger.debug(
            "unable to get ASGI scope for route observation. args: %s, kwargs: %s",
            args,
            kwargs,
        )
        return

    if route := asgi_scope.get("route"):
        # This is the common FastAPI case. Starlette won't set the route key.
        context.signature = common.build_signature(route.name, route.endpoint)
        context.path_template = route.path
    elif endpoint := asgi_scope.get("endpoint"):
        if isinstance(endpoint, StaticFiles):
            context.signature = f"StaticFiles(directory={endpoint.directory})"
        elif inspect.isclass(endpoint) and issubclass(endpoint, HTTPEndpoint):
            context.signature = common.build_signature(endpoint.__name__, endpoint)
        elif callable(endpoint) and (func_name := getattr(endpoint, "__name__", None)):
            context.signature = common.build_signature(func_name, endpoint)
    else:
        logger.debug("WARNING: did not find endpoint for starlette route observation")
        return

    logger.debug(
        "Found starlette route",
        signature=context.signature,
        path_template=context.path_template,
    )


class ContrastSessionDictProxy(wrapt.ObjectProxy):
    """
    Custom ObjectProxy we use to wrap dicts returned by starlette's request.session
    property. These proxied dicts have a trigger for trust-boundary-violation on
    __setitem__.
    """

    def __setitem__(self, key, value):
        result = None
        try:
            result = self.__wrapped__.__setitem__(key, value)
        finally:
            analyze_setitem(result, (self, key, value))

        return result


@fail_quietly("Failed to analyze session dict __setitem__")
def analyze_setitem(result, args):
    policy = registry.get_policy_by_name("starlette.sessions.dict.__setitem__")
    analyze(policy, result, args, {})


def build_session_patch(orig_prop, patch_policy):
    def session_fget(*args, **kwargs):
        """
        Function used to replace fget for starlette's request.session property.
        This function returns proxied dictionaries - see ContrastSessionDictProxy.
        """
        session_dict = orig_prop.fget(*args, **kwargs)

        context = contrast.REQUEST_CONTEXT.get()
        if context is None:
            return session_dict

        return ContrastSessionDictProxy(session_dict)

    return property(session_fget, orig_prop.fset, orig_prop.fdel)


safe_cs__apply_source = fail_quietly()(cs__apply_source)


def build_stream_patch(orig_func, patch_policy):
    assert len(patch_policy.source_nodes) == 1
    source_node = patch_policy.source_nodes[0]

    async def stream_patch(wrapped, instance, args, kwargs):
        """
        Make starlette.requests.Request.stream a source. Policy patches are not equipped
        to handle generator functions, so we're using a custom monkeypatch here.

        Conveniently, the original generator does not have a return value (it only
        yields) and does not expect any values from `send()` / `asend()`, which
        simplifies this implementation significantly.
        """
        if not ((context := contrast.REQUEST_CONTEXT.get()) and context.assess_enabled):
            async for item in wrapped(*args, **kwargs):
                yield item
            return

        async for item in wrapped(*args, **kwargs):
            safe_cs__apply_source(
                context,
                source_node,
                item,
                instance,
                item,
                args,
                kwargs,
            )
            yield item

    return wrap_and_watermark(orig_func, stream_patch)


def patch_starlette_requests(starlette_requests_module):
    build_and_apply_patch(
        starlette_requests_module.Request, "session", build_session_patch
    )
    build_and_apply_patch(
        starlette_requests_module.Request, "stream", build_stream_patch
    )


def patch_starlette_routing(starlette_routing_module):
    build_and_apply_patch(
        starlette_routing_module.Router, "__call__", build___call___patch
    )


def reverse_patches():
    unregister_module_patcher(STARLETTE_REQUESTS)
    starlette_routing_module = sys.modules.get(STARLETTE_ROUTING)
    if starlette_routing_module:
        patch_manager.reverse_patches_by_owner(starlette_routing_module.Router)

    unregister_module_patcher(STARLETTE_REQUESTS)
    starlette_requests_module = sys.modules.get(STARLETTE_REQUESTS)
    if starlette_requests_module:
        patch_manager.reverse_patches_by_owner(starlette_requests_module.Request)


def register_patches():
    register_module_patcher(patch_starlette_requests, STARLETTE_REQUESTS)
    register_module_patcher(patch_starlette_routing, STARLETTE_ROUTING)
